__all__ = (
    # 'get_audio_volume'
    'get_battery_info',
    'get_cpu_temp',
    'get_cpu_usage',
    'get_datetime',
    'get_disk_usage',
    'get_host',
    'get_hostname',
    'get_mem_usage',
    'get_net_stats',
    'get_uptime',
)


import os
import re
import shlex
import time
from asyncio import subprocess as aiosp
from datetime import datetime
from string import Formatter

from .errors import InvalidArgError
from .formatting import ElapsedTime, ConditionalFormatStr, format_uptime
from .utils import join_options
from ._types import (
    POWERS_OF_1024,
    Contents,
    FormatStr,
    MetricSymbol,
    NmConnFilterSpec,
    StorageMeasure,
)

from collections.abc import Callable, Iterable
from typing import Literal, TypeAlias, NamedTuple, Any
from enum import Enum


# Check if all field functions will work:
try:
    os.stat('/proc/stat')
except PermissionError:
    raise CompatibilityWarning(
        "Field functions which require access to procfs will break"
        " on this system."
    )
else:
    import psutil


async def get_audio_volume(
    fmt: FormatStr = "{:02.0f}{state}",
    *args, **kwargs
) -> Contents:
    '''
    Currently awaiting a more practical implementation that supports
    instantaneous updates. A socket would be quite nice for that.
    '''
##    '''Returns system audio volume from ALSA amixer. SIGUSR1 is used to
##    notify the thread of volume changes from button presses.'''
    pat = re.compile(r'.*\[(\d+)%\] \[(\w+)\]')

    cmd = await aiosp.create_subprocess_shell(
        "amixer sget Master",
        stdout=aiosp.PIPE
    )
    out, err = await cmd.communicate()
    state_list = (l.strip() for l in reversed(out.decode().splitlines()))

    pcts, states = zip(*(
        m.groups()
        for l in state_list
        if (m := pat.match(l))
    ))
    avg_pct = sum(int(p) for p in pcts) // len(pcts)
    is_on = any(s == 'on' for s in states)
    # return is_on, avg_pct
    state = "" if is_on else "M"
    return fmt.format(avg_pct, state=state)


##def ctx_get_battery_info(
##    fmt: FormatStr = "{icon}{pct:02.0f}",
##    # fmt: FormatStr = "{icon}{pct:02.0f}{state}",
##    *args, **kwargs
##) -> Context:
##    '''
##    Battery capacity as a percent and whether or not it is charging.
##    '''
##
##    # if not (battery := psutil.sensors_battery()):
##    battery = psutil.sensors_battery()
##    if not battery:
##        return Context()
####        # return (None, None)
##    # state = "CHG" if battery.power_plugged else ''
##
##    # Progressive/dynamic battery icons!
##    icon_bank = "    ".split()
##    # return icon_bank
##
##    def mapper(n):
##        icon = ""
##        for i, test in enumerate((10, 25, 50, 75, 100)):
##            if n <= test:
##                icon = icon_bank[i]
##                return icon + " "
##
##    if battery.power_plugged:
##        icon = ""
##        # state = "CHG"
##    else:
##        icon = mapper(battery.percent)
##        # state = ""
##    # print(repricon)
##
##    info = fmt.format_map({'icon': icon, 'pct': battery.percent, 'state': battery.power_plugged})
####    return battery.power_plugged, info
##    return info


async def get_battery_info(
    fmt: FormatStr = "{pct:02.0f}{state}",
    *args, **kwargs
) -> Contents:
    '''
    Battery capacity as a percent and whether or not it is charging.

    :param fmt: A curly-brace format string with\
        two optional named fields:
            - ``pct``: Current battery percent as a :class:`float`
            - ``state``: Whether or not the battery is charging

        Defaults to ``"{pct:02.0f}{state}"``
    :type fmt: :class:`FormatStr`
    '''
    battery = psutil.sensors_battery()
    if not battery:
        return ""
    state = "CHG" if battery.power_plugged else ''
    info = fmt.format_map({'pct': battery.percent, 'state': state})
    return info


def get_cpu_temp(
    fmt: str = "{temp:02.0f}{scale}",
    in_fahrenheit=False,
    *args, **kwargs
) -> Contents:
    '''
    Current CPU temperature in Celcius or Fahrenheit.

    :param fmt: A curly-brace format string with\
        two optional named fields:
            - ``temp``: Current CPU temperature as a :class:`float`
            - ``scale``: ``'C'`` or ``'F'``, depending on `in_fahrenheit`

        Defaults to ``"{temp:02.0f}{scale}"``
    :type fmt: :class:`FormatStr`

    :param in_fahrenheit: Display the temperature in Fahrenheit instead
        of Celcius, defaults to ``False``
    :type in_fahrenheit: :class:`bool`
    '''
    scales = ('C', 'F')
    scale = scales[in_fahrenheit]

    temps = psutil.sensors_temperatures(in_fahrenheit)
    match temps:
        case {'k10temp': t} | {'coretemp': t}:
            current = t[0].current
        case _:
            current = '??'
    return fmt.format_map({'temp': current, 'scale': scale})


def get_cpu_usage(
    fmt: FormatStr = "{:02.0f}%",
    interval: float = None,
    *args, **kwargs
) -> Contents:
    '''
    System CPU usage in percent over a specified interval.

    :param fmt: A curly-brace format string with one positional field,
        defaults to ``"{:02.0f}%"``
    :type fmt: :class:`FormatStr`

    :param interval: Time to block before returning a result.
        Only set this in a threaded :class:`Field`.
    :type interval: :class:`float`, optional
    '''
    return fmt.format(psutil.cpu_percent(interval))


def get_datetime(
    fmt: str = "%Y-%m-%d %H:%M:%S",
    *args, **kwargs
) -> Contents:
    '''
    Current time as formatted with `fmt`.

    :param fmt: A format string with %-based format codes used by
        ``datetime.strftime()``, defaults to ``"%Y-%m-%d %H:%M:%S"``
    :type fmt: :class:`str`

    .. seealso:: `Datetime strftime() format codes\
    <https://docs.python.org/3/library/datetime.html#strftime-and-strptime\
    -format-codes>`_

    '''
    return datetime.now().strftime(fmt)


async def get_disk_usage(
    fmt: FormatStr = "{free:.1f}{unit}",
    path: os.PathLike = '/',
    unit: MetricSymbol = 'G',
    *args, **kwargs
) -> Contents:
    '''
    Disk usage of a partition at a given path.

    :param fmt: A curly-brace format string with\
        five optional named fields:
            - ``unit``: The same value as `unit`
            - ``total``: Total disk partition size
            - ``used``: Used disk space
            - ``free``: Free disk space (``total - used``)
            - ``percent``: Percent of total disk space that is used\
                (``used / total``, a :class:`float`)

        Defaults to ``"{free:.1f}{unit}"``
    :type fmt: :class:`FormatStr`

    :param path: The path to a directory or file on a disk partition,
        defaults to ``'/'``
    :type path: :class:`os.PathLike`

    :param unit: The unit prefix symbol, defaults to ``'G'``
    :type unit: :class:`MetricSymbol`

    :raises: :exc:`errors.InvalidArgError`
        if `unit` is not a valid :class:`MetricSymbol`
    '''
    if unit not in POWERS_OF_1024:
        raise InvalidArgError(
            f"Invalid unit: {unit!r}\n"
            f"'unit' must be one of "
            f"{join_options(POWERS_OF_1024)}."
        )

    disk = psutil.disk_usage(path)
    factor = POWERS_OF_1024[unit]
    converted = {
        meas: (val if meas == 'percent' else val/factor)
        for meas, val in disk._asdict().items()
    }
    converted['unit'] = unit
    usage = fmt.format_map(converted)
    return usage


def get_host(
    fmt: FormatStr = "{nodename}",
    *args, **kwargs
) -> Contents:
    '''
    System host information using :func:`os.uname()`.

    :param fmt: A curly-brace format string with\
        five optional named fields:
            - ``nodename``: System hostname
            - ``sysname``: Operating system kernel name
            - ``release``: Kernel release
            - ``version``: Kernel version
            - ``machine``: Machine architecture

        Defaults to ``"{nodename}"``
    :type fmt: :class:`FormatStr`
    '''
    keys = HostOption.__args__
    host = fmt.format_map(dict(zip(keys, os.uname())))
    return host


def get_hostname(*args, **kwargs) -> Contents:
    '''System hostname.'''
    return os.uname().nodename


def get_mem_usage(
    fmt: FormatStr = "{used:.1f}{unit}",
    unit: MetricSymbol = 'G',
    *args, **kwargs
) -> Contents:
    '''
    System memory usage.

    :param fmt: A curly-brace format string with the following optional\
        named fields:
            - ``unit``: The same value as `unit`
            - ``available``: Available memory
            - ``total``: Total physical memory excluding swap
            - ``used``: Memory in use
            - ``percent``: Percent memory usage calculated as\
                ``(total - available) / total * 100``

        Defaults to ``"{free:.1f}{unit}"``
        See the `psutil documentation\
            <https://psutil.readthedocs.io/en/latest/#psutil\
            .virtual_memory>`_ for all possible optional named fields.
    :type fmt: :class:`FormatStr`

    :param unit: The unit prefix symbol, defaults to ``'G'``
    :type unit: :class:`MetricSymbol`

    :raises: :exc:`errors.InvalidArgError`
        if `unit` is not a valid :class:`MetricSymbol`
    '''
    if unit not in POWERS_OF_1024:
        raise InvalidArgError(
            f"Invalid unit: {unit!r}\n"
            f"'unit' must be one of "
            f"{join_options(POWERS_OF_1024)}."
        )

    memory = psutil.virtual_memory()
    factor = POWERS_OF_1024[unit]
    converted = {
        meas: (val if meas == 'percent' else val/factor)
        for meas, val in memory._asdict().items()
    }
    converted['unit'] = unit
    usage = fmt.format_map(converted)
    return usage


async def get_net_stats(
    # device: str = None,
    fmt: FormatStr = "{name}",
    nm: bool = True,
    nm_filt: NmConnFilterSpec = None,
    default: str = "",
    *args, **kwargs
) -> Contents:
    '''
    Active network info from either `NetworkManager` or `iwconfig`.

    :param fmt: A curly-brace format string with\
        five optional named fields:
            - ``name``: The connection name
            - ``uuid``: The connection uuid
            - ``type``: The connection type
            - ``device``: The connection device

        Defaults to ``"{name}"``
    :type fmt: :class:`FormatStr`

    :param nm: Use `NetworkManager`, defaults to ``True``
    :type nm: :class:`bool`

    :param nm_filt: Filter from active `NetworkManager` connections
    :type nm_filt: :class:`NmConnFilterSpec`, optional

    :param default: The string to replace `fmt` fields when there is no active connection.
    :type default: :class:`str`
    '''
    # If the user has NetworkManager, get the active connections:
    if nm:

        if nm_filt:
            # We use --terse for easier parsing.
            cmd = f"nmcli --terse connection show {shlex.join(nm_filt.items())}"
        else:
            cmd = "nmcli --terse connection show --active"

        proc = await aiosp.create_subprocess_shell(
            cmd,
            stdout=aiosp.PIPE,
            stderr=aiosp.PIPE
        )
        out, err = await proc.communicate()
        results = out.decode().splitlines()

        # Split in reverse to avoid butchering names with colons:
        conns = (c.rsplit(':', 3) for c in results)

        # By default, nmcli sorts connections by type:
        #   ethernet, wifi, tun, ...
        # We only need the first result.

        match next(conns, None):

            case None:
                keynames = ('name', 'uuid', 'typ', 'device')
                return fmt.format_map({key: default for key in keynames})

            case (name, uuid, typ, device):
                profile = {
                    # The command output duplicates all backslashes to
                    # escape them and adds more before colons in names.
                    # Take special care to preserve literal \ and : in
                    # original names while removing extra backslashes:
                    'name': name.replace('\\\\', '\\').replace('\\:', ':'),
                    'uuid': uuid,
                    'type': typ,
                    'device': device,
                }
                return fmt.format_map(profile)

            case _:
                return default

    else:
        cmd = await aiosp.create_subprocess_shell(
            "iwconfig",
            stdout=aiosp.PIPE
        )
        out, err = await cmd.communicate()
        if_list = out.decode().splitlines()

        ssid = next(
            line.split(':"')[1].strip('" ')
            for line in if_list
            if line.find("SSID")
        )
        return ssid


def get_player(
    service: str | None = None,
    fmt: FormatStr = "{artist} - {title}",
    max_artists: int = 1,
    limit: int = 40,
    # setupvars = None,
) -> Contents:
    '''
    Playerctl data.

    :param fmt: A curly-brace format string with eleven optional named
        fields:
        ``trackid``, ``length``, ``artUrl``, ``album``, ``albumArtist``,
        ``artist``, ``autoRating``, ``discNumber``, ``title``,
        ``trackNumber``, and ``url``.
    :type fmt: :class:`FormatStr`
    '''
    # from gi.repositories import GLib
    # glib_loop = GLib.MainLoop()
    # players = setupvars['players']

    import pympris
    import dbus
    from dbus.mainloop.glib import DBusGMainLoop

    dbus_loop = DBusGMainLoop()
    bus = dbus.SessionBus(mainloop=dbus_loop)

    player_ids = list(pympris.available_players())
    type ServiceName = str
    players: dict[ServiceName] = {
        (p := pympris.MediaPlayer(i, bus)).root.Identity: p
        for i in player_ids
    }
    # service = 'cmus'
    if service is None and players:
        # Just choose one at random:
        p = players.popitem()[1]
    elif service not in players:
        return ""
    else:
        p = players[service]
    metadata = {
        key.split(':', 1)[-1]: (
            join_options(val, limit=max_artists) if not isinstance(val, str)
            and isinstance(val, Iterable)
            else val
        )
        for key, val in p.player.Metadata.items()
    }
    contents = fmt.format_map(metadata)
    # print(contents)
    if limit and len(contents) > limit:
        return contents[:limit] + "..."
    return contents


async def get_uptime(
    fmt: FormatStr,
    dynamic: bool = True,
    sep: str = ':',
    setupvars = None,
    *args, **kwargs
) -> Contents:
    '''
    System uptime.

    This function does some neat things.
    When `dynamic` is ``True``, the format string is split into groups
    using `sep`. If any format string fields in a group evaluate to
    0, the whole group is hidden.

    For example, using the format string ``"{days}d:{hours}h:{mins}m"``,
    86580 seconds is represented as ``1d:0h:3m`` when `dynamic` is
    ``False``, but would be shortened to ``1d:3m`` when `dynamic`
    is ``True``.

    Using only one unit, like ``"{hours}"``, will round `hours` to an
    integer by default. If floating point values are desired, a format
    spec must be used:
    ``"{hours:.4f}"``

    Format specs can be used to great effect, like in all other Field
    functions which accept `fmt`:
    ``"{hours:_^10.4f}"``

    .. seealso:: The Python `format spec mini language documentation
        <https://docs.python.org/3/library/string.html#format-\
        specification-mini-language>`_

    :param fmt: A curly-brace format string with seven optional named
        fields: ``secs``, ``mins``, ``hours``, ``days``, ``weeks``,
        ``months``, and ``years``

    :type fmt: :class:`FormatStr`

    :param dynamic: Given `sep`, automatically hide groups of units when
        they are ``0``, defaults to ``True``
    :type dynamic: :class:`bool`

    :param sep: Delimits groups of format fields to be hidden/shown
        together, defaults to ``":"``
    :type sep: :class:`str`
    '''
    secs = time.time() - psutil.boot_time()

    if not setupvars:
        conditional = ConditionalFormatStr(fmt, sep)

        setupvars = {
            'fnames': conditional.fnames,
            'groups': conditional.groups,
            'sep': sep,
        }
    lookup_table = ElapsedTime.in_desired_units(secs, setupvars['fnames'])
    setupvars['namespace'] = lookup_table

    if dynamic:
        out = format_uptime(secs, **setupvars)
        return out

    return fmt.format_map(setupvars['namespace'])




#TODO: Implement dynamic icons!
#TODO: Finish Mocp line!


__all__ = ('Bar', 'BarConfig')


import asyncio
import json
import os
import sys
import threading
import time
from copy import deepcopy

import scuff
from scuff.tools import ScuffText

from .constants import (
    CLEAR_LINE,
    CONFIG_FILE,
    CSI,
    DEBUG,
    HIDE_CURSOR,
    UNHIDE_CURSOR,
)

from . import cli
from . import field_funcs
from . import utils
from .errors import *
from .field import Field, FieldPrecursor
from .formatting import FmtStrStructure, FormatterFieldSig
from .namespaces import BarConfigSpec, BarSpec, FieldSpec, _CmdOptionSpec
from ._types import (
    Args,
    ASCII_Icon,
    ASCII_Separator,
    Bar,
    BarConfig,
    ConsoleControlCode,
    FieldName,
    FieldOrder,
    FileContents,
    FormatStr,
    Icon,
    JSONText,
    Kwargs,
    Line,
    Pattern,
    Separator,
    Unicode_Icon,
    Unicode_Separator,
)

from collections.abc import Iterable, Iterator, Mapping, Sequence
from os import PathLike
from typing import (
    IO,
    NoReturn,
    Required,
    Self,
)


class BarConfig(dict):
    '''
    Build and transport configs between files, dicts and command lines.

    :param options: Optional :class:`namespaces.BarConfigSpec` parameters
        that override those of `defaults`
    :type options: :class:`namespaces.BarConfigSpec`

    :param defaults: Parameters to use by default,
        defaults to :attr:`Bar._default_params`
    :type defaults: :class:`Mapping`

    .. note:: `options` and `defaults` must be :class:`Mapping` instances
        of form :class:`namespaces.BarConfigSpec`

    '''
    def __init__(
        self,
        options: BarConfigSpec = {},
        defaults: BarConfigSpec = None,
    ) -> None:
        if defaults is None:
            self.defaults = Bar._default_params.copy()
        else:
            self.defaults = defaults.copy()
        self.options = options.copy()

        # This is why we subclass dict:
        self.update(self.defaults | self.options)
        self._remove_unserializable()
        self._unify_field_defs()
        self.file = None
        self.file_contents = None
        # debug = options.get('debug', None) or DEBUG

    def __repr__(self) -> str:
        cls = type(self).__name__
        file = self.file
        maybe_file = f"{file=}, " if file else ""
        params = self.copy()
        return f"<{cls} {maybe_file}{params}>"

    @classmethod
    def from_file(
        cls,
        file: PathLike = None,
        *,
        defaults: BarConfigSpec = None,
        overrides: BarConfigSpec = {},
    ) -> BarConfig:
        '''
        Return a new :class:`BarConfig` from a config file path.

        :param file: The filepath to the config file,
            defaults to ``'~/.mybar.json'``
        :type file: :class:`PathLike`

        :param defaults: The base :class:`namespaces.BarConfigSpec` dict
            whose params the new :class:`BarConfig` will override,
            defaults to :attr:`Bar._default_params`
        :type defaults: :class:`namespaces.BarConfigSpec`

        :param overrides: Additional param overrides to the config file
        :type overrides: :class:`namespaces.BarConfigSpec`

        :returns: A new :class:`BarConfig` instance
        :rtype: :class:`BarConfig`
        :raises: :exc:`OSError` for issues with accessing the file
        '''
        if defaults is None:
            defaults = Bar._default_params
        overrides = overrides.copy()

        if file is None:
            file = overrides.pop('config_file', CONFIG_FILE)
        is_json = (os.path.splitext(file)[1] == '.json')
        reader = (cls._read_file, cls._read_json)[is_json]
        try:
            from_file, text = reader(file)
        except OSError as e:
            raise e.with_traceback(None)

        options = utils.nested_update(from_file, overrides)

        config = cls(options, defaults)
        config.file = file
        config.file_contents = text
        return config

    @classmethod
    def validate(cls, ) -> bool | NoReturn:
        '''
        Check if the config file has errors.
        '''

    def _unify_field_defs(self) -> Self:
        '''
        Gather loose Field definitions and join them with any others
        found in a `field_definitions`, if it exists.
        Return the updated :class:`BarConfig` with its new consolidated
        `field_definitions`.

        :returns: The processed config
        :rtype: :class:`namespaces.BarConfigSpec`
        '''
        all_except_defs = (
            BarConfigSpec.__optional_keys__
            | BarConfigSpec.__required_keys__
        )
        defs = self.pop('field_definitions', {})
        config = self.copy()
        for param in tuple(self):
            if param not in all_except_defs:
                field_def = self.pop(param)
                if not isinstance(field_def, Mapping):
                    raise ValueError(
                        f"Invalid config file option: {field_def!r}"
                    )
                # The name of the Field defintion overrides this option:
                field_def.pop('name', None)
                defs[param] = field_def

        # Remove command line options:
        for option in _CmdOptionSpec.__optional_keys__:
            defs.pop(option, None)


        self['field_definitions'] = defs
        return self

    def write_with_approval(self, file: PathLike = None) -> bool:
        '''
        Write a config file to `file`.
        Prompt the user for approval before writing.

        :param file: Write to this file, defaults to :obj:`CONFIG_FILE`
        :type file: :class:`PathLike`

        :returns: ``True`` if the file was written, ``False`` if not
        '''
        if file is None:
            file = CONFIG_FILE

        approved = cli.FileManager._get_write_new_approval(
            file,
            dft_choice=True
        )
        is_json = (os.path.splitext(file)[1] == '.json')
        writer = (self._write_file, self._write_json)[is_json]
        if approved:
            writer(file)
            print(f"Wrote new config file to {file!r}")
            return True

        return False

    @classmethod
    def _read_file(cls, file: PathLike) -> tuple[BarSpec, FileContents]:
        '''
        Read a given config file.
        Parse and convert its contents to a dict.
        Return the dict along with the raw text of the file.

        :param file: The file to convert
        :type file: :class:`PathLike`

        :returns: The converted file and its raw text contents
        :rtype: tuple[
            :class:`namespaces.BarSpec`,
            :class:`FileContents`
            ]
        '''
        absolute = os.path.abspath(os.path.expanduser(file))
        p = scuff.FileParser(file=absolute)
        from_file = p.to_py()
        data = cls._unify_field_defs(from_file)
        text = p._string
        return data, text

    def _write_file(self, file: PathLike, indent: int = 4) -> None:
        '''
        Write :class:`BarConfig` params to a config file.

        :param file: The file to write to
        :type file: :class:`PathLike`

        :param indent: How many spaces to indent by, defaults to ``4``
        :type indent: :class:`int`
        '''
        absolute = os.path.abspath(os.path.expanduser(file))
        if absolute == CONFIG_FILE and not os.path.exists(absolute):
            cli.FileManager._maybe_make_config_dir()
        with open(absolute, 'w') as f:
            f.write(self.as_scuff() + '\n')

    @classmethod
    def _read_json(cls, file: PathLike) -> tuple[BarConfigSpec, JSONText]:
        '''
        Read a given JSON config file.
        Convert its contents to a dict and return it along with the
        raw text of the file.

        :param file: The file to convert
        :type file: :class:`PathLike`

        :returns: The converted file and its raw text
        :rtype: tuple[:class:`namespaces.BarConfigSpec`,
            :class:`JSONText`]
        '''
        absolute = os.path.abspath(os.path.expanduser(file))
        with open(absolute, 'r') as f:
            from_file = json.load(f)
            text = f.read()
        data = cls._unify_field_defs(from_file)
        return data, text

    def _write_json(self, file: PathLike, indent: int = 4) -> None:
        '''
        Write the :class:`BarConfig` params to a JSON file.

        :param file: The file to write to
        :type file: :class:`PathLike`

        :param indent: How many spaces to indent by, defaults to ``4``
        :type indent: :class:`int`
        '''
        absolute = os.path.abspath(os.path.expanduser(file))
        if absolute == CONFIG_FILE and not os.path.exists(absolute):
            cli.FileManager._maybe_make_config_dir()
        with open(absolute, 'w') as f:
            f.write(json.dumps(self, indent=indent) + '\n')

    def _remove_unserializable(self) -> Self:
        '''
        Remove Python-specific API elements like functions which cannot
        be serialized for writing to config files.
        '''
        self.pop('config_file', None)

        fields = self.pop('field_definitions', {})
        # Throw away unneeded Field implementation details:
        newfields = {}
        for name, params in fields.items():
            newfields[name] = {
                param: val for param, val in params.items()
                if param not in FieldSpec.unserializable
            }
        if newfields:
            self['field_definitions'] = newfields

        return self

    def as_scuff(self) -> ScuffText:
        '''
        Return a :class:`BarConfig` as a string with Scuff formatting.
        '''
        scuffer = scuff.PyParser(unquoted=tuple(Field._default_fields))
        return scuffer.to_scuff(self)

    def as_json(self, indent: int = 4) -> JSONText:
        '''
        Return a :class:`BarConfig` as a string with JSON formatting.

        :param indent: How many spaces to indent by, defaults to ``4``
        :type indent: :class:`int`
        '''
        return json.dumps(self, indent=indent)


class Bar:
    '''
    Create highly customizable status bars.

    :param fields: An iterable of default field names or :class:`Field`
        instances, defaults to ``None``
    :type fields: Iterable[:class:`FieldPrecursor`]

    :param template: A curly-brace format string with field names,
        defaults to ``None``
    :type template: :class:`FormatStr`

    :param separator: The field separator string when `fields` is given,
        or a sequence of 2 of these, defaults to ``'|'``
        The first string is used in terminal environments where
        only ASCII is supported.
        The second string is used in graphical environments where
        support for Unicode is more likely.
        This enables the same :class:`Bar` instance to use the most
        optimal separator automatically.
    :type separator: :class:`Separator` |
        Sequence[:class:`ASCII_Separator`,
        :class:`Unicode_Separator`], optional

    :param refresh: How often in seconds the bar automatically prints
        itself, defaults to ``1.0``
    :type refresh: :class:`float`

    :param count: Only print the bar this many times, or never stop
        when ``None``, defaults to ``None``
    :type count: :class:`int` | ``None``

    :param break_lines: Print each refresh after a newline character
        (``'\\n'``), defaults to False
    :type break_lines: :class:`bool`

    :param clock_align: Whether to synchronize redraws at the start of
        each new second (updates to the clock are shown accurately),
        defaults to ``True``
    :type clock_align: :class:`bool`

    :param join_empty_fields: Whether to draw separators around fields
        with no content, defaults to ``False``
    :type join_empty_fields: :class:`bool`

    :param override_cooldown: Cooldown in seconds between handling
        sequential field overrides, defaults to ``1/60``.
        A longer cooldown typically means less flickering.
    :type override_cooldown: :class:`float`

    :param thread_cooldown: How long a field thread loop sleeps after
        checking if the bar is still running, defaults to ``1/8``.
        Between executions, unlike async fields, a threaded field sleeps
        for several iterations of `thread_cooldown` seconds that always
        add up to :attr:`Field.interval` seconds.
        Between sleeps, it checks if the bar has stopped.
        A shorter cooldown means more chances to check if the bar has
        stopped and a faster exit time.
    :type thread_cooldown: :class:`float`

    :param unicode: Use Unicode variants of `separator` and Field
        icons, if given; optional
    :type unicode: :class:`bool`

    :param stream: The bar's output stream,
        defaults to :attr:`sys.stdout`
    :type stream: :class:`IO`

    :raises: :exc:`errors.InvalidOutputStreamError` when `stream` does
        not implement the IO protocol
    :raises: :exc:`errors.IncompatibleArgsError` when
        neither `template` nor `fields` are given
    :raises: :exc:`errors.IncompatibleArgsError` when
        `template` is ``None``
        but no `separator` given
    :raises: :exc:`TypeError` when `fields` is not iterable, or when
        `template` is not a string
    '''

    _default_field_order = [
        'hostname',
        'uptime',
        'cpu_usage',
        'cpu_temp',
        'mem_usage',
        'disk_usage',
        'battery',
        'net_stats',
        'datetime',
    ]

    _default_params = {
        'separator': '|',
        'refresh': 1.0,
        'field_order': list(_default_field_order)
    }

    def __init__(
        self,
        template: FormatStr = None,
        fields: Iterable[FieldPrecursor] = None,
        *,
        field_order: Iterable[FieldName] = None,
        separator: Separator | Sequence[ASCII_Separator,
                                        Unicode_Separator] = '|',
        refresh: float = 1.0,
        count: int | None = None,
        break_lines: bool = False,
        clock_align: bool = True,
        join_empty_fields: bool = False,
        override_cooldown: float = 1/8,
        thread_cooldown: float = 1/8,
        unicode: bool = True,
        stream: IO = sys.stdout,
        debug: bool = DEBUG,  # Not yet implemented!
    ) -> None:
        # Ensure the output stream has the required methods:
        self._check_stream(stream)
        self._stream = stream

        # Check all the required parameters.
        if fields is None:

            if template is None:
                raise IncompatibleArgsError(
                    f"Either a list of Fields `fields` "
                    f"or a format string `template` is required."
                )

            elif not isinstance(template, str):
                raise TypeError(
                    f"Format string `template` must be a string, "
                    f"not {type(template)}"
                )

            # Good to use template:
            else:
                parsed = FmtStrStructure.from_str(template)
                parsed.validate_fields(Field._default_fields, True, True)
                # names = parsed.get_names()
                fields = parsed

        # Fall back to using fields.
        elif not isinstance(separator, Iterable):
            raise TypeError("The `fields` argument must be iterable.")

        if unicode is None:
            unicode = not self.stream.isatty()
        self._unicode = unicode

        if isinstance(separator, str):
            separator = (separator, separator)
        elif not isinstance(separator, Sequence):
            raise TypeError("`separator` must be a sequence")
        self._separators = separator

        field_names, fields = self._normalize_fields(fields)

        # Preserve custom field order, enabling duplicates:
        if field_order is None:
            field_order = field_names
        self._fields = fields
        self._field_order = field_order
        self._buffers = dict.fromkeys(self._fields, '')

        self.template = template

        self._endline = '\n' if break_lines is True else '\r'
        self.join_empty_fields = join_empty_fields
        self.refresh_rate = refresh

        self.count = count
        self._print_countdown = count

        self.clock_align = clock_align
        self._override_queue = asyncio.Queue(maxsize=1)
        self._override_cooldown = override_cooldown
        self._thread_cooldown = thread_cooldown

        # The dict mapping Field names to the threads running them.
        # The bar must join each before printing a line with ``count=1``:
        self._threads = {}
        self._printer_thread = None
        self._printer_loop = None
        self._funcs = {}

        self._coros = {}
        self._timely_fields = []

        # Calling run() sets this Event. Unsetting it stops all fields:
        self._can_run = threading.Event()
        self._running = self._can_run.is_set

        # The bar's async event loop:
        # self._thread_loop = asyncio.new_event_loop()
        self._loop = asyncio.new_event_loop()

        self._file = None
        self._config = None

    def __contains__(self, field: FieldPrecursor) -> bool:
        if isinstance(other, str):
            weak_test = (field in self._field_order)
            return weak_test
        elif isinstance(field, Field):
            less_weak_test = (field.name in self._field_order)
            return less_weak_test
        else:
            return False

    def __eq__(self, other: Self) -> bool:
        if not all(
            getattr(self, attr) == getattr(other, attr)
            for attr in (
                '_fields',
                '_field_order',
                '_separators',
                'join_empty_fields',
            )
        ):
            return False
        if self.template == other.template:
            return True
        return False

    def __iter__(self) -> Iterator:
        # Include duplicates.
        return iter(self._fields[name] for name in self._field_order)

    def __len__(self) -> int:
        # Include duplicates.
        return len(self._field_order)

    def __repr__(self) -> str:
        names = self._field_order
        fields = utils.join_options(names, final_sep='', limit=3)
        cls = type(self).__name__
        return f"{cls}(fields=[{fields}])"

    @property
    def config(self) -> BarConfig:
        '''
        The :class:`BarConfig` used to instantiate the Bar, if applicable.
        '''
        return self._config

    @property
    def file(self) -> PathLike:
        '''
        The config file used to instantiate the Bar, if applicable.
        '''
        return self._file

    @classmethod
    def from_config(
        cls,
        config: BarConfig,
        *,
        overrides: BarSpec = {},
        ignore_with: Pattern | tuple[Pattern] | None = '//'
    ) -> Self:
        '''
        Return a new :class:`Bar` from a :class:`BarConfig`
        or a dict of :class:`Bar` parameters.
        Ignore keys and list elements starting with `ignore_with`,
        which is ``'//'`` by default.
        If :param ignore_with: is ``None``, do not remove any values.

        :param config: The :class:`Mapping` to convert
        :type config: :class:`Mapping`

        :param overrides: Replace items in `config` with these params,
            defaults to ``{}``
        :type overrides: :class:`namespaces.BarSpec`

        :param ignore_with: A pattern to ignore, defaults to ``'//'``
        :type ignore_with: :class:`Pattern` |
            tuple[:class:`Pattern`] | ``None``, optional

        :returns: A new :class:`Bar`
        :rtype: :class:`Bar`

        :raises: :exc:`errors.IncompatibleArgsError` when
            no `field_order` is defined
            or when no `template` is defined
        :raises: :exc:`errors.DefaultFieldNotFoundError` when either
            a field in `field_order` or
            a field in `template`
            cannot be found in :attr:`Field._default_fields`
        :raises: :exc:`errors.UndefinedFieldError` when
            a custom field name in `field_order` or
            a custom field name in `template`
            is not properly defined in `field_definitions`
        :raises: :exc:`errors.InvalidFieldSpecError` when
            a field definition is not of the form
            :class:`namespaces.FieldSpec`

        .. note:: `config` can be a regular :class:`Mapping`
            as long as it matches the form of :class:`namespaces.BarSpec`.

        '''
        if ignore_with is None:
            data = deepcopy(config)
        else:
            data = utils.scrub_comments(config, ignore_with)
        data = utils.nested_update(data, overrides)
        bar_params = utils.nested_update(cls._default_params, data)

        field_order = bar_params.pop('field_order', None)
        field_icons = bar_params.pop('field_icons', {})
        field_defs = bar_params.pop('field_definitions', {})
        template = bar_params.get('template')
        if field_icons is None:
            field_icons = {}

        if template is None:
            if not field_order:
                # field_order = list(field_icons)
                raise IncompatibleArgsError(
                    "A bar format string `template` is required"
                    " when field order list `field_order` is undefined"
                    " or empty."
                )
        else:
            parsed = FmtStrStructure.from_str(template)
            parsed.validate_fields(Field._default_fields, True, True)
            field_order = parsed.get_names()

        # Gather Field parameters and instantiate new Fields:
        fields = {}
        expctd_name = (
            "the name of a default or"
            " properly defined `custom` Field in field list"
        )

        for name in field_order:
            field_params = field_defs.get(name)

            cust_icons = field_icons.get(name, None)
            if cust_icons is not None:
                if isinstance(cust_icons, str):
                    # When one icon is given, override both defaults:
                    cust_icons = (cust_icons, cust_icons)

            match field_params:
                case None:
                    # The field is strictly default.
                    if name in fields:
                        continue

                    if cust_icons is not None:
                        overrides = {'icon': cust_icons}
                    else:
                        overrides = {}

                    try:
                        field = Field.from_default(name, overrides=overrides)
                    except DefaultFieldNotFoundError:
                        exc = utils.make_error_message(
                            DefaultFieldNotFoundError,
                            blame=f"{name!r}",
                            expected=expctd_name
                        )
                        raise exc from None

                case {'custom': True}:
                    # The field is custom, so it is only defined in
                    # 'field_definitions' and will not be found as a default.
                    name = field_params.pop('name', name)
                    del field_params['custom']
                    field_params['icon'] = cust_icons
                    field = Field(**field_params, name=name)

                case {}:
                    # The field is a default overridden by the user.
                    if cust_icons is not None:
                        field_params['icon'] = cust_icons
                    try:
                        field = Field.from_default(
                            name,
                            overrides=field_params
                        )
                    except DefaultFieldNotFoundError:
                        exc = utils.make_error_message(
                            UndefinedFieldError,
                            doing_what=f"parsing {name!r} definition",
                            blame=f"{name!r}",
                            expected=expctd_name,
                            epilogue=(
                                f"(In config files, remember to set"
                                f" `custom=true`"
                                f" for custom field definitions.)"
                            ),
                        )
                        raise exc from None

                case _:
                    # Edge cases.
                    exc = utils.make_error_message(
                        InvalidFieldSpecError,
                        doing_what=f"parsing {name!r} definition",
                        details=(
                            f"Invalid Field specification: {field_params!r}",
                        ),
                        indent_level=1
                    )
                    raise exc from None

            fields[name] = field

        bar = cls(
            fields=tuple(fields.values()),
            field_order=field_order,
            **bar_params
        )
        bar._config = config
        if hasattr(config, 'file'):
            bar._file = config.file
        return bar

    @classmethod
    def from_file(
        cls,
        file: PathLike = None,
        overrides: BarSpec = {}
    ) -> Self:
        '''
        Generate a new :class:`Bar` by reading a config file.

        :param file: The config file to read,
            defaults to :obj:`constants.CONFIG_FILE`
        :type file: :class:`PathLike`

        :param overrides: Replace items in `config` with these params,
            defaults to ``{}``
        :type overrides: :class:`namespaces.BarSpec`

        :returns: A new :class:`Bar`
        :rtype: :class:`Bar`
        '''
        config = BarConfig.from_file(file)
        bar = cls.from_config(config, overrides=overrides)
        bar._file = config.file
        return bar

    @property
    def clearline_char(self) -> str:
        '''
        A special character printed to TTY streams between refreshes.
        Its purpose is to clear characters left behind by longer lines.
        '''
        if self._stream is None:
            return None
        if self._stream.isatty() and self._endline == '\r':
            return CLEAR_LINE + self._endline
        else:
            return ''
        return clearline

    @property
    def fields(self) -> tuple[Field]:
        '''A tuple of the bar's Field objects.'''
        return tuple(self)

    @property
    def in_a_tty(self) -> bool:
        '''
        ``True`` if the bar was run from a terminal, otherwise ``False``.
        '''
        if self._stream is None:
            return False
        return self._stream.isatty()

    @property
    def running(self) -> bool:
        '''
        ``True`` if the bar is currently running, otherwise ``False``.
        '''
        return self._running()

    @property
    def separator(self) -> Separator:
        '''
        The field separator as determined by the output stream.
        Defaults to the TTY sep (self._separators[0]) if no stream is set.
        '''
        if self._stream is None:
            # Default to using the ASCII separator:
            return self._separators[0]
        return self._separators[self._unicode]

    def append(self, field: FieldPrecursor) -> Self:
        '''
        Append a new Field to the bar.

        `field` will be converted to :class:`Field`,
        appended to the field list and joined to the end of the bar.
        If :attr:`Bar.template` is defined,
        it will override the new field order.

        :param field: The field to append
        :type field: :class:`FieldPrecursor`

        :returns: The modified bar
        :rtype: :class:`Bar`
        '''
        (name,), normalized = self._normalize_fields((field,))
        self._fields.update(normalized)
        self._buffers[name] = ''
        self._field_order.append(name)
        return self

    def extend(self, fields: Iterable[FieldPrecursor]) -> Self:
        '''
        Append several new Fields to the bar.
        Field precursors in `fields` will be converted to :class:`Field`,
        appended to the field list and joined to the end of the bar.
        If :attr:`Bar.template` is defined,
        it will override the new field order.
        the fields are displayed.

        :param fields: The fields to append
        :type fields: Iterable[:class:`FieldPrecursor`]

        :returns: The modified bar
        :rtype: :class:`Bar`
        '''
        #TODO: Consider Bar.template += FormatterFieldSig(field).as_string()
        names, normalized = self._normalize_fields(fields)
        self._fields.update(normalized)
        self._buffers.update(dict.fromkeys(names, ''))
        self._field_order.extend(names)
        return self

    @staticmethod
    def _check_stream(
        stream: IO,
        raise_on_fail: bool = True
    ) -> NoReturn | None:
        '''
        Check if an IO stream has proper methods for use by :class:`Bar`.
        If any methods are missing,
            return ``False`` if `raise_on_fail` is ``False``, or
            raise :exc:`InvalidBarError` if `raise_on_fail` is ``True``.
        Otherwise, return ``True``.

        :param stream: The IO stream object to test
        :type stream: :class:`IO`

        :param raise_on_fail: Raise exception if `stream` fails the check,
            defaults to ``True``
        :type raise_on_fail: :class:`bool`

        :raises: :exc:`InvalidOutputStreamError` if `raise_on_fail` is
            ``True`` and the stream fails the test and lacks required
            instance methods
        '''
        io_methods = ('write', 'flush', 'isatty')
        if not all(hasattr(stream, m) for m in io_methods):
            if not raise_on_fail:
                return False
            io_method_calls = [m + '()' for m in io_methods]
            joined = utils.join_options(io_method_calls, final_sep='and')
            raise InvalidOutputStreamError(
                f"Output stream {stream!r} needs {joined} methods."
            ) from None
        return True

    def _normalize_fields(
        self,
        fields: Iterable[FieldPrecursor],
    ) -> tuple[FieldOrder, dict[FieldName, Field]]:
        '''
        Convert :class:`Field` precursors to :class:`Field` instances.

        Take an iterable of field names, :class:`Field` instances, or
        :class:`formatting.FormatterFieldSig` tuples and convert them to
        corresponding default Fields.
        Return a dict mapping field names to Fields, along with a
        duplicate-preserving list of field names that were found.

        :param fields: An iterable of :class:`FieldPrecursor` to convert
        :type fields: :class:`Iterable[FieldPrecursor]`

        :returns: The field order and a dict mapping field names to
            :class:`Field` instances
        :rtype: :class:`tuple[FieldOrder, Mapping[FieldName, Field]]`

        :raises: :exc:`errors.InvalidFieldError` when an element
            of `fields` is not a :class:`FieldPrecursor`
        '''
        normalized = {}
        names = []

        for field in fields:
            match field:
                case Field():
                    new_field = field
                    new_field._bar = self
                case str(name):
                    new_field = Field.from_default(
                        name=field,
                        overrides={'bar': self},
                    )
                case FormatterFieldSig(name=None):
                        # Skip sigs that lack fields:
                        continue
                case FormatterFieldSig():
                    new_field = Field.from_default(
                        name=field.name,
                        overrides={'bar': self}
                    )
                    new_field._fmtsig = field.as_string()
                case _:
                    msg = f"Invalid field precursor: {field!r}"
                    raise InvalidFieldError(msg)

            normalized[new_field.name] = new_field
            names.append(new_field.name)

        return names, normalized

    def run(
        self,
        once: bool = None,
        stream: IO = None,
        *,
        bg: bool = False
    ) -> None:
        '''
        Run the bar in the specified output stream.
        Block until an exception is raised and exit smoothly.

        :param stream: The IO stream in which to run the bar,
            defaults to :attr:`Bar._stream`
        :type stream: :class:`IO`

        :param once: Run the bar only once,
            defaults to internal state
        :type once: :class:`bool`

        :param bg: (Not fully implemented)
            Run the bar in the background without printing,
            defaults to ``False``
            This is used for :meth:`Bar.line_generator` and for testing.
        :type bg: :class:`bool`

        :returns: ``None``

        .. note::
            This method cannot be called within an async event loop.
        '''
        if self.count == 0:
            return

        try:
            if stream is not None:
                self._stream = stream
            if once is None:
                once = False if self.count is None else (self.count <= 1)
            elif once:
                self.count = 1

            # Allow the bar to run repeatedly in the same interpreter:
            if self._loop.is_closed():
                self._loop = asyncio.new_event_loop()

            self._prepare_fields()
            self._can_run.set()
            for thread in tuple(self._threads.values()):
                thread.start()
            self._preload_timely_fields()

            if not once:
                self._start_printer()

            # This blocks:
            self._loop.run_until_complete(self._gather_coros())

            if once:
                # Wait for threads to finish to get a full line, then print.
                while self._threads:
                    time.sleep(self._thread_cooldown)
                self._print_one_line()
                self._can_run.clear()

            # Prevent the bar from exiting when there are no coroutines:
            while self._running():
                time.sleep(self._thread_cooldown)

        except KeyboardInterrupt:
            pass

        finally:
            self._shutdown()

    def _prepare_fields(self) -> None:
        once = False if self.count is None else (self.count <= 1)
        overriding = False
        for field in self:
            if field.overrides_refresh:
                overriding = True

            if field.threaded:
                thread = field.make_thread(bar=self, run_once=once)
                self._threads[thread.name] = thread
            elif field.timely:
                self._timely_fields.append(field)
            else:
                self._coros[field.name] = field.run(bar=self, once=once)

##        if overriding:  # Currently disabled!
##            self._coros.append(self._handle_overrides())

    async def _gather_coros(self) -> None:
        '''
        Run :class:`Bar` coroutines in parallel.
        '''
        await asyncio.gather(*self._coros.values())

    def _preload_timely_fields(self) -> None:
        '''
        Update the bar buffers for Fields with :attr:`Field.timely`.
        This is used most often right before printing a line.
        '''
        for f in self._timely_fields:
            if f.is_async:
                result = self._loop.run_until_complete(
                    f._func(*f.args, **f.kwargs)
                )
            else:
                result = f._func(*f.args, **f.kwargs)

            if f.template is not None:
                contents = f.template.format(result, icon=f.icon)
            else:
                if f.always_show_icon or result:
                    contents = f.icon + result
                else:
                    contents = result
            self._buffers[f.name] = contents

    def _start_printer(self) -> None:
        '''
        Make a thread in which to run
            :meth:`Bar._threaded_continuous_line_printer` and start it.
        '''
        thread_name = 'PRINTER'
        self._printer_thread = threading.Thread(
            target=self._threaded_continuous_line_printer,
            name=thread_name,
        )
        self._threads[thread_name] = self._printer_thread
        self._printer_thread.start()

    def _threaded_continuous_line_printer(self) -> None:
        '''
        The bar's primary line-printing mechanism.
        Fields with the ``timely`` attribute get run here.
        They run once every refresh,
        effectively ignoring their ``interval`` attributes.
        Other fields are responsible for sending data to bar buffers.
        This only writes using the current buffer contents.
        '''
        # Make an event loop for this thread:
        self._printer_loop = asyncio.new_event_loop()

        # Flushing the buffer before writing to it fixes poor i3bar
        # alignment.
        self._stream.flush()

        beginning = self.clearline_char
        if self.in_a_tty:
            self._stream.write(CSI + HIDE_CURSOR)

        # Print something right away just so that the bar is not empty,
        # but not if the print count matters:
        if self.count is None:
            line = self._make_one_line()
            self._stream.write(beginning + line + self._endline)
            self._stream.flush()

        using_format_str = (self.template is not None)
        sep = self.separator
        clock = time.monotonic
        step = self._thread_cooldown
        needed = round(self.refresh_rate / step)
        count = 0
        first_cycle = True

        if self.clock_align:
            # Sleep until the beginning of the next second.
            clock = time.time
            time.sleep(1 - (clock() % 1))

        start_time = clock()
        while self._running():

            # Sleep until the next refresh cycle in little steps to
            # check if the bar has stopped.
            if count < needed and not first_cycle:
                count += 1

                # Latency from nonzero execution times causes drift,
                # where sleeps become out-of-sync and the bar skips
                # field updates.
                # This is especially noticeable in fields that update
                # routinely, such as the time.

                # To negate drift, when sleeping until the beginning of
                # the next cycle, we must also compensate for latency.
                time.sleep(
                    step - (
                        # Get the current latency, which can vary:
                        clock() % step
                        # (clock() - start_time) % step Preserves offset
                    )
                )
                continue

            count = 0
            if first_cycle:
                first_cycle = False

            # Run time-sensitive fields right away:
            for f in self._timely_fields:
                if f.is_async:
                    result = self._printer_loop.run_until_complete(
                        f._func(*f.args, **f.kwargs)
                    )
                else:
                    result = f._func(*f.args, **f.kwargs)

                if f.template is not None:
                    contents = f.template.format(result, icon=f.icon)
                else:
                    if f.always_show_icon or result:
                        contents = f.icon + result
                    else:
                        contents = result
                self._buffers[f.name] = contents

            if using_format_str:
                line = self.template.format_map(self._buffers)
            else:
                line = sep.join(
                    buf for field in self._field_order
                        if (buf := self._buffers[field])
                        or self.join_empty_fields
                )

            self._stream.write(beginning + line + self._endline)
            self._stream.flush()

            if self.count:
                self._print_countdown -= 1
                if self._print_countdown == 0:
                    self._can_run.clear()
                    self._printer_loop.stop()
                    self._printer_loop.close()
                    return

        self._printer_loop.stop()
        self._printer_loop.close()

    def _shutdown(self) -> None:
        '''
        Notify fields that the bar has stopped and join threads.
        Also print a newline and unhide the cursor if the bar was
        running in a terminal.
        '''
        self._can_run.clear()

        threads = tuple(self._threads.items())
        for thread_name, thread in threads:
            thread.join()
            self._threads.pop(thread_name)

        if self.in_a_tty:
            self._stream.write('\n')
            self._stream.write(CSI + UNHIDE_CURSOR)

    async def _handle_overrides(self) -> None:
        '''
        Print a line when fields with overrides_refresh send new data.
        '''
        sep = self.separator
        using_format_str = (self.template is not None)
        beginning = self.clearline_char

        start_time = time.time()
        while self._running():

            try:
                # Wait until a field with overrides_refresh sends new
                # data to be printed:
                field, contents = await self._override_queue.get()

            except RuntimeError:
                # asyncio raises RuntimeError if the event loop closes
                # while queue.get() is waiting for a value.
                return

            if using_format_str:
                line = self.template.format_map(self._buffers)
            else:
                line = sep.join(
                    buf for field in self._field_order
                        if (buf := self._buffers[field])
                        or self.join_empty_fields
                )

            self._stream.write(beginning + line + self._endline)
            self._stream.flush()
            await asyncio.sleep(self._override_cooldown)

    def _make_one_line(self) -> Line:
        '''Make a line using the bar's field buffers.
        This method is not meant to be called from within a loop.
        '''
        if self.template is not None:
            line = self.template.format_map(self._buffers)
        else:
            line = self.separator.join(
                buf for field in self._field_order
                    if (buf := self._buffers[field])
                    or self.join_empty_fields
            )
        return line

    def _print_one_line(self) -> None:
        '''Print a line to the buffer stream only once.
        This method is not meant to be called from within a loop.
        '''
        beginning = self.clearline_char

        if self._timely_fields:
            self._preload_timely_fields()
        self._stream.write(beginning + self._make_one_line() + self._endline)

        self._stream.flush()

    def line_generator(self) -> Iterator:
        '''
        Return a generator yielding status lines.
        This requires running the Fields in a separate thread,
        which has yet to be implemented.
        '''
        while self._running():
            if self._timely_fields:
                self._preload_timely_fields()
            line = self._make_one_line()
            yield line

    def current_line(self) -> Line:
        '''
        Return the most recently printed line.
        '''
        return self._make_one_line()




__all__ = (
    'logger'
)


import logging
import os.path

from .constants import CONFIG_FILE


logging.basicConfig(
    level='DEBUG',
    filename=os.path.dirname(CONFIG_FILE) + '/mybar.log',
    filemode='w',
    datefmt='%Y-%m-%d_%H:%M:%S.%f',
    format='[{asctime}] ({levelname}:{name}) {message}',
    style='{',
)
logger = logging.getLogger(__name__)




#TODO: Implement dynamic icons!
#TODO: Finish Mocp line!


__all__ = (
    'Field',
)


import asyncio
import asyncio.subprocess as aiosp
import os.path
import shlex
import threading
import time
from os import PathLike

from . import field_funcs
from . import _setups
from . import utils
from .constants import DEBUG
from .errors import *
from .formatting import FormatterFieldSig
from .namespaces import FieldSpec
from ._types import (
    Args,
    ASCII_Icon,
    Bar,
    Contents,
    Field,
    FieldFunc,
    FieldFuncSetup,
    FieldName,
    FormatStr,
    Icon,
    Kwargs,
    Unicode_Icon,
)

from collections.abc import Callable, Sequence
from typing import (
    NamedTuple,
    Never,
    Required,
    Self,
    TypeAlias,
    TypeVar
)


class Field:
    '''
    Continuously generate and format one bit of information in a
    :class:`Bar`.

    Pre-defined default Fields can be looked up by name using
    :meth:`Field.from_default`.

    :param name: A unique identifier for the new Field,
        defaults to `func`.:attr:`__name__`
    :type name: :class:`str`

    :param func: The Python function to run at every `interval` if no
        `constant_output` is set
    :type func: :class:`FieldFunc`

    :param icon: The Field icon (or a sequence of one for ASCII terminals
        and one for terminals that support Unicode fonts) positioned in
        front of Field contents or in place of ``{icon}`` in `template`,
        if provided, defaults to ``''``
    :type icon: :class:`Icon` | `Sequence[ASCII_Icon, Unicode_Icon]`

    :param template: A curly-brace format string. This parameter is
        **required** if `icon` is ``None``.

        Valid placeholders:
            - ``{icon}`` references `icon`
            - ``{}`` references Field contents

        Example:
            | When the Field's current contents are ``'69F'`` and its\
            icon is ``'TEMP'``,
            | ``template='[{icon}]: {}'`` shows as ``'[TEMP]: 69F'``

    :type template: :class:`FormatStr`

    :param interval: How often in seconds Field contents are updated,
        defaults to ``1.0``
    :type interval: :class:`float`

    :param clock_align: Update contents at the start of each second,
        defaults to ``False``
    :type clock_align: :class:`bool`

    :param timely: Run `func` as soon as possible after every refresh,
        defaults to ``False``
    :type timely: :class:`bool`

    :param overrides_refresh: Ensure updates to this Field re-print the
        Bar between refreshes, defaults to ``False``
    :type overrides_refresh: :class:`bool`

    :param threaded: Run this Field in a separate thread,
        defaults to ``False``
    :type threaded: :class:`bool`

    :param always_show_icon: Show icons even when contents are empty,
        defaults to ``False``
    :type always_show_icon: :class:`bool`

    :param run_once: Permanently set contents by running `func` only once,
        defaults to ``False``
    :type run_once: :class:`bool`

    :param constant_output: Permanently set contents instead of running
        a function
    :type constant_output: :class:`str`, optional

    :param bar: Attach the :class:`Field` to this :class:`Bar`
    :type bar: :class:`Bar`, optional

    :param args: Positional args passed to `func`
    :type args: :class:`Args`, optional

    :param kwargs: Keyword args passed to `func`
    :type kwargs: :class:`Kwargs`, optional

    :param setup: A special callback that updates `kwargs` with static
        data that `func` would otherwise have to evaluate every time it
        runs
    :type setup: :class:`FieldFuncSetup`, optional

    :param command: A shell command to run
    :type command: :class:`str`

    :param script: The path to a shell script to run
    :type script: :class:`PathLike`

    :param allow_multiline: Don't join the output of a command or script
        if it spans multiple lines, defaults to ``False``
    :type allow_multiline: :class:`bool`

    :raises: :exc:`errors.IncompatibleArgsError` when
        neither `func` nor `constant_output` are given
    :raises: :exc:`errors.IncompatibleArgsError` when
        neither `icon` nor `template` are given
    :raises: :exc:`TypeError` when `func` is not callable
    :raises: :exc:`TypeError` when `setup`, if given, is not callable
    '''
    _default_fields = {

        'hostname': {
            'name': 'hostname',
            'func': field_funcs.get_hostname,
            'run_once': True
        },

        'host': {
            'name': 'host',
            'func': field_funcs.get_host,
            'kwargs': {
                'fmt': "{nodename}",
            },
            'run_once': True
        },

        'uptime': {
            'name': 'uptime',
            'func': field_funcs.get_uptime,
            'kwargs': {
                'fmt': '{days}d:{hours}h:{mins}m',
                'dynamic': True,
                'sep': ':',
            },
            'setup': _setups.setup_uptime,
            'timely': True,
            'clock_align': True,
            'icon': 'Up ',
        },

        'cpu_usage': {
            'name': 'cpu_usage',
            'func': field_funcs.get_cpu_usage,
            'kwargs': {
                'fmt': "{:02.0f}%",
            },
            'interval': 5,
            'threaded': True,
            'icon': 'CPU ',
        },

        'cpu_temp': {
            'name': 'cpu_temp',
            'func': field_funcs.get_cpu_temp,
            'kwargs': {
                'fmt': "{temp:02.0f}{scale}",
            },
            'interval': 5,
            'threaded': True,
        },

        'mem_usage': {
            'name': 'mem_usage',
            'func': field_funcs.get_mem_usage,
            'kwargs': {
                'fmt': "{used:.1f}{unit}",
                'unit': 'G',
            },
            'interval': 5,
            'icon': 'Mem ',
        },

        'disk_usage': {
            'name': 'disk_usage',
            'func': field_funcs.get_disk_usage,
            'kwargs': {
                'fmt': "{free:.1f}{unit}",
                'path': '/',
                'unit': 'G',
            },
            'interval': 5,
            'icon': '/:',
        },

        'battery': {
            'name': 'battery',
            'kwargs': {
                'fmt': "{pct:02.0f}{state}",
            },
            'func': field_funcs.get_battery_info,
            'threaded': True,
            'icon': 'Bat ',
        },

        'net_stats': {
            'name': 'net_stats',
            'func': field_funcs.get_net_stats,
            'kwargs': {
                # 'device': None,
                'fmt': "{name}",
                'nm': True,
                'nm_filt': None,
                'default': "",
            },
            'interval': 5,
        },

        'datetime': {
            'name': 'datetime',
            'func': field_funcs.get_datetime,
            'kwargs': {
                'fmt': "%Y-%m-%d %H:%M:%S",
            },
            'clock_align': True,
            'timely': True,
        }

    }

    def __init__(
        self,
        *,
        name: FieldName = None,
        func: FieldFunc = None,
        icon: Icon | Sequence[ASCII_Icon, Unicode_Icon] = '',
        template: FormatStr = None,
        interval: float = 1.0,
        clock_align: bool = False,
        timely: bool = False,
        overrides_refresh: bool = False,
        threaded: bool = False,
        always_show_icon: bool = False,
        run_once: bool = False,
        constant_output: str = None,
        bar: Bar = None,
        args: Args = None,
        kwargs: Kwargs = None,
        setup: FieldFuncSetup = None,
        command: str = None,
        script: PathLike = None,
        allow_multiline: bool = False,
    ) -> None:
        if all(_ is None for _ in (func, script, command, constant_output)):
            #NOTE: This will change when dynamic icons and templates are
            # implemented.
            raise IncompatibleArgsError(
                "Either a function `func` that returns a string, a"
                " shell script `script`, a shell command `command` or"
                " a `constant_output` string is required."
            )

        if script is not None:
            script = os.path.abspath(script)
            command = self._fetch_script(script)
            if name is None:
                name = os.path.basename(script)
        if command is not None:
            func = self._run_command
            if name is None:
                name = command.split(None, 1)[0]
        elif not callable(func) and constant_output is None:
            raise TypeError(
                f"Type of `func` must be callable, not {type(func)}"
            )
        elif name is None:
            name = func.__name__

        self.name = name
        self._func = func
        self.args = () if args is None else args
        self.kwargs = {} if kwargs is None else kwargs

        if setup is not None and not callable(setup):
            # `setup` was given but is of the wrong type.
            raise TypeError(
                f"Type of 'setup' must be callable, not {type(setup)}"
            )
        self._setupfunc = setup

        if isinstance(icon, str):
            icon = (icon, icon)
        elif not icon:
            icon = ('', '')
        elif not isinstance(icon, Sequence):
            raise TypeError("`icon` must be a sequence")
        self._icons = icon

        if template is None and icon is None:
            raise IncompatibleArgsError(
                "An icon is required when `template` is None."
            )
        self.template = template

        self.is_async = asyncio.iscoroutinefunction(func)
        if self.is_async or threaded or timely:
            self._callback = func
        else:
            # Wrap synchronous functions if they aren't special:
            self._callback = self._asyncify

        self.allow_multiline = allow_multiline
        self.always_show_icon = always_show_icon
        self._bar = bar
        self._buffer = None
        self.clock_align = clock_align
        self.command = command
        self.constant_output = constant_output
        self.interval = interval
        self.overrides_refresh = overrides_refresh
        self.run_once = run_once
        self.script = script
        self.threaded = threaded
        self.timely = timely

        self._do_setup()

    def __repr__(self) -> str:
        cls = type(self).__name__
        name = self.name
        # attrs = utils.join_options(...)
        return f"{cls}({name=})"

    def __eq__(self, other: Self) -> bool:
        if not all(
            getattr(self, attr) == getattr(other, attr)
            for attr in (
                'clock_align',
                'always_show_icon',
                'constant_output',
                'template',
                '_icons',
            )
        ):
            return False

        if self._func is None:
            if self.constant_output == other.constant_output:
                return True
        elif (
            self._func is other._func
            and self.args == other.args
            and self.kwargs == other.kwargs
        ):
            return True
        return False

    @classmethod
    def from_default(
        cls: Self,
        name: str,
        *,
        overrides: FieldSpec = {},
        source: dict[FieldName, FieldSpec] = None,
        fmt_sig: FormatStr = None
    ) -> Self:
        '''Quickly get a default Field and customize its parameters.

        :param name: Name of the default :class:`Field` to access or customize
        :type name: :class:`str`

        :param overrides: Custom parameters that override those of the default Field
        :type overrides: :class:`namespaces.FieldSpec`, optional

        :param source: The :class:`dict` in which to look up default fields,
            defaults to :attr:`Field._default_fields`
        :type source: dict[:class:`FieldName`, :class:`namespaces.FieldSpec`]

        :returns: A new :class:`Field`
        :rtype: :class:`Field`
        :raises: :exc:`errors.DefaultFieldNotFoundError` when `source` does not contain `name`
        '''
        if source is None:
            source = cls._default_fields

        try:
            default: FieldSpec = source[name].copy()
        except KeyError:
            raise DefaultFieldNotFoundError(
                f"{name!r} is not the name of a default Field."
            ) from None

        if overrides.get('interval', None) != default.get('interval', None):
            overrides['timely'] = False
        if 'kwargs' in overrides and 'kwargs' in default:
            overrides['kwargs'] = default['kwargs'] | overrides['kwargs']

        spec = default | overrides
        field = cls(**spec)
        field._fmt_sig = fmt_sig
        return field

    @classmethod
    def from_format_string(cls, fmt: FormatStr) -> Self:
        '''
        Get a :class:`Field` from a curly-brace field in a format string.

        :param template: The format string to convert
        :type template: :class`FormatStr`
        '''
        sig = FormatterFieldSig.from_str(fmt)
        field = cls.from_default(sig.name)
        field._fmtsig = fmt
        return field

    @property
    def icon(self) -> str:
        '''The field icon as determined by the output stream of its bar.
        It defaults to the TTY icon (:attr:`self._icon[1]`) if no bar is set.
        '''
        if self._bar is None:
            # Default to using the ASCII icon:
            return self._icons[0]
        unicode = self._bar._unicode or not self._bar._stream.isatty()
        return self._icons[unicode]

    async def _asyncify(self, *args, **kwargs) -> Contents:
        '''Wrap a synchronous function in a coroutine for simplicity.'''
        return self._func(*args, **kwargs)

    def _fetch_script(self, script: PathLike) -> str:
        with open(script, 'r') as f:
            script = f.read()
        return script

    async def _run_command(self, *args, **kwargs) -> Contents:
        cmd = shlex.join(shlex.split(self.command))
        proc = await aiosp.create_subprocess_shell(
            cmd,
            stdout=aiosp.PIPE,
            stderr=aiosp.PIPE
        )
        out, err = await proc.communicate()
        result = out.decode()
        if not self.allow_multiline:
            result = ' '.join(result.splitlines())
        return result

    @staticmethod
    def _format_contents(
        text: str,
        icon: str,
        template: FormatStr = None,
        always_show_icon: bool = False
    ) -> Contents:
        '''A helper function that formats field contents.'''
        if template is None:
            if always_show_icon or text:
                return icon + text
            else:
                return text
        else:
            return template.format(text, icon=icon)

    def _auto_format(self, text: str) -> Contents:
        '''Non-staticmethod _format_contents...'''
        if self.template is None:
            if self.always_show_icon or text:
                return self.icon + text
            else:
                return text
        else:
            return self.template.format(text, icon=self.icon)

    def _do_setup(self, args: Args = None, kwargs: Kwargs = None) -> None:
        '''
        Use the pre-defined _setupfunc() to gather constant variables
        for func() which might only be evaluated at runtime.
        '''
        if self._setupfunc is None:
            return

        if args is None:
            args = self.args
        if kwargs is None:
            kwargs = self.kwargs

        try:
            if asyncio.iscoroutinefunction(self._setupfunc):
                setupvars = asyncio.get_event_loop().run_until_complete(
                    self._setupfunc(*self.args, **self.kwargs)
                )
            else:
                setupvars = self._setupfunc(*self.args, **self.kwargs)

        # If _setupfunc raises FailedSetup with a backup value,
        # use it as the field's new constant_output:
        except FailedSetup as e:
            backup = e.backup
            self.constant_output = self._auto_format(backup)
            return

        # On success, give new values to kwargs to pass to _func().
        # return setupvars
        kwargs['setupvars'] = setupvars

    def sync_run(self, once: bool = None) -> Contents:
        # Do not run fields which have a constant output;
        # only set their bar buffer.
        if self.constant_output is not None:
            return self._auto_format(self.constant_output)
        if self.is_async:
            result = asyncio.get_event_loop().run_until_complete(
                self._func(*self.args, **self.kwargs)
            )
        else:
            result = self._func(*self.args, **self.kwargs)
        contents = self._auto_format(result)
        return contents

    def gen_run(self, once: bool = None) -> Contents:
        # Do not run fields which have a constant output;
        # only set their bar buffer.
        if self.constant_output is not None:
            return self._auto_format(self.constant_output)
        while True:
            if self.is_async:
                result = self._func(*self.args, **self.kwargs)
            else:
                result = asyncio.get_event_loop().run_until_complete(
                    self._func(*self.args, **self.kwargs)
                )
            contents = self._auto_format(result)
            yield contents

    async def run(self, bar: Bar, once: bool) -> None:
        '''
        Run an async :class:`Field` callback and send its output to a
        status bar.

        :param bar: Send results to this status bar
        :type bar: :class:`Bar`

        :param once: Run the Field function once
        :type once: :class:`bool`
        '''
        self._check_bar(bar)
        # Do not run fields which have a constant output;
        # only set their bar buffer.
        if self.constant_output is not None:
            contents = self._format_contents(
                self.constant_output,
                self.icon,
                self.template,
                self.always_show_icon
            )
            bar._buffers[self.name] = contents
            return

        func = self._callback
        running = bar._can_run.is_set
        clock = time.monotonic
        using_format_str = (self.template is not None)
        last_val = None

        # Use the pre-defined _setupfunc() to gather constant variables
        # for func() which might only be evaluated at runtime:
        if self._setupfunc is not None:
            try:
                if asyncio.iscoroutinefunction(self._setupfunc):
                    setupvars = (
                        await self._setupfunc(*self.args, **self.kwargs)
                    )
                else:
                    setupvars = self._setupfunc(*self.args, **self.kwargs)

            # If _setupfunc raises FailedSetup with a backup value,
            # use it as the field's new constant_output and update the
            # bar's buffer:
            except FailedSetup as e:
                backup = e.backup
                contents = self._format_contents(
                    backup,
                    self.icon,
                    self.template,
                    self.always_show_icon
                )
                self.constant_output = contents
                bar._buffers[self.name] = str(contents)
                return

            # On success, give new values to kwargs to pass to func().
            self.kwargs['setupvars'] = setupvars

        # Run at least once at the start to ensure bars have contents:
        result = await func(*self.args, **self.kwargs)
        last_val = result
        contents = self._auto_format(result)
        bar._buffers[self.name] = contents

        if self.run_once or once:
            return

        if self.clock_align:
            # Sleep until the beginning of the next second.
            clock = time.time
            await asyncio.sleep(1 - (clock() % 1))
        start_time = clock()

        # The main loop:
        while running():

            if bar.count:
                # Stop before running a cycle that could last too long:
                if (
                    bar._print_countdown == 0
                    or self.interval > (
                        bar._print_countdown * bar.refresh_rate 
                    )
                ):
                    bar._coros.pop(self.name, None)
                    return

            result = await func(*self.args, **self.kwargs)
            # Latency from nonzero execution times causes drift, where
            # sleeps become out-of-sync and the bar skips field updates.
            # This is especially noticeable in fields that update
            # routinely, such as the time.

            # To negate drift, when sleeping until the beginning of the
            # next cycle, we must also compensate for latency.
            await asyncio.sleep(
                # Time until next refresh:
                self.interval - (
                    # Get the current latency, which can vary:
                    clock() % self.interval
                    # (clock() - start_time) % self.interval  # Preserve offset
                )
            )

            if result == last_val:
                continue
            last_val = result

            if using_format_str:
                contents = self.template.format(result, icon=self.icon)
            else:
                if self.always_show_icon or result:
                    contents = self.icon + result
                else:
                    contents = result

            bar._buffers[self.name] = contents

            # Send new field contents to the bar's override queue and
            # print a new line between refresh cycles.
            if self.overrides_refresh:
                try:
                    bar._override_queue.put_nowait((self.name, contents))

                except asyncio.QueueFull:
                    # Since the bar buffer was just updated, do nothing
                    # if the queue is full. The update may still show
                    # while the queue handles the current override. If
                    # not, the line will update at the next refresh cycle.
                    pass

    def run_threaded(self, bar: Bar, once: bool) -> None:
        '''
        Run a blocking :class:`Field` func and send its output to a
        status bar.

        :param bar: Send results to this status bar
        :type bar: :class:`Bar`

        :param once: Run the Field function once
        :type once: :class:`bool`
        '''
        self._check_bar(bar)

        # Do not run fields which have a constant output;
        # only set their bar buffer.
        if self.constant_output is not None:
            contents = self._format_contents(
                self.constant_output,
                self.icon,
                self.template,
                self.always_show_icon
            )
            bar._buffers[self.name] = contents
            return

        func = self._callback
        running = bar._can_run.is_set
        clock = time.monotonic
        using_format_str = (self.template is not None)
        last_val = None

        # If the field's callback is asynchronous,
        # it must be run in a new event loop.
        local_loop = asyncio.new_event_loop()

        # Use the pre-defined _setupfunc() to gather constant variables
        # for func() which might only be evaluated at runtime:
        if self._setupfunc is not None:
            try:
                if asyncio.iscoroutinefunction(self._setupfunc):
                    setupvars = local_loop.run_until_complete(
                        self._setupfunc(*self.args, **self.kwargs)
                    )
                else:
                    setupvars = self._setupfunc(*self.args, **self.kwargs)

            # If _setupfunc raises FailedSetup with a backup value,
            # use it as the field's new constant_output and update the
            # bar buffer:
            except FailedSetup as e:
                backup = e.args[0]
                contents = self._format_contents(
                    backup,
                    self.icon,
                    self.template,
                    self.always_show_icon
                )
                self.constant_output = contents
                bar._buffers[self.name] = str(contents)
                return

            # On success, give new values to kwargs to pass to func().
            self.kwargs['setupvars'] = setupvars

        # Run at least once at the start to ensure bar is not empty:
        if self.is_async:
            result = local_loop.run_until_complete(
                func(*self.args, **self.kwargs)
            )
        else:
            result = func(*self.args, **self.kwargs)
        last_val = result
        contents = self._format_contents(
            result,
            self.icon,
            self.template,
            self.always_show_icon
        )
        bar._buffers[self.name] = contents

        if self.run_once or once:
            local_loop.stop()
            local_loop.close()
            bar._threads.pop(self.name)  #NOTE: Eventually, id(self)
            return

        step = bar._thread_cooldown
        needed = round(self.interval / step)
        count = 0
        first_cycle = True

        if self.clock_align:
            # Sleep until the beginning of the next second.
            clock = time.time
            time.sleep(1 - (clock() % 1))

        start_time = clock()
        while running():

            if bar.count:
                # Stop before running a cycle that could last too long:
                if (
                    bar._print_countdown == 0
                    or self.interval > (
                        bar._print_countdown * bar.refresh_rate 
                    )
                ):
                    local_loop.stop()
                    local_loop.close()
                    bar._coros.pop(self.name, None)
                    return


            # Sleep until the next refresh cycle in little steps to
            # check if the bar has stopped.
            if count < needed and not first_cycle:
                count += 1

                # Latency from nonzero execution times causes drift,
                # where sleeps become out-of-sync and the bar skips
                # field updates.
                # This is especially noticeable in fields that update
                # routinely, such as the time.

                # To negate drift, when sleeping until the beginning of
                # the next cycle, we must also compensate for latency.
                time.sleep(
                    step - (
                        # Get the current latency, which can vary:
                        clock() % step
                        # (clock() - start_time) % step To preserve offset
                    )
                )
                continue

            count = 0
            if first_cycle:
                first_cycle = False

            if self.is_async:
                result = local_loop.run_until_complete(
                    func(*self.args, **self.kwargs)
                )
            else:
                result = func(*self.args, **self.kwargs)

            if result == last_val:
                continue
            last_val = result

            if using_format_str:
                contents = self.template.format(result, icon=self.icon)
            else:
                if self.always_show_icon or result:
                    contents = self.icon + result
                else:
                    contents = result

            bar._buffers[self.name] = contents

            # Send new field contents to the bar's override queue and print a
            # new line between refresh cycles.
            if self.overrides_refresh:
                try:
                    bar._override_queue._loop.call_soon_threadsafe(
                        bar._override_queue.put_nowait, (self.name, contents)
                    )
                except asyncio.QueueFull:
                    # Since the bar buffer was just updated, do nothing
                    # if the queue is full. The update may still show
                    # while the queue handles the current override. If
                    # not, the line will update at the next refresh cycle.
                    pass

        local_loop.stop()
        local_loop.close()

    @staticmethod
    def _check_bar(
        bar: Bar,
        raise_on_fail: bool = True
    ) -> Never | bool:
        '''
        Check if a bar has the right attributes to use a run() function.
        If these attributes are missing,
            return ``False`` if `raise_on_fail` is ``True``,
            or raise :exc:`InvalidBarError` if `raise_on_fail` is ``False``.

        :param bar: The status bar object to test
        :type bar: :class:`Bar`

        :param raise_on_fail: Raise an exception if `bar` fails the check,
            defaults to ``False``
        :type raise_on_fail: :class:`bool`

        :raises: :exc:`InvalidBarError` if the stream fails the test
            and lacks required attributes
        '''
        required_attrs = (
            '_buffers',
            '_can_run',
            '_override_queue',
            '_stream',
        )
        if not all(hasattr(bar, a) for a in required_attrs):
            if not raise_on_fail:
                return False
            raise InvalidBarError(
                f"Status bar {bar!r} is missing certain attributes"
                f" required to run `Field.run()`"
                f" and `Field.run_threaded()`."
            ) from None
        return True

    def make_thread(self, bar: Bar, run_once: bool) -> None:
        '''
        Return a thread that runs the :class:`Field`'s callback.

        :param bar: Send results to this status bar
        :type bar: :class:`Bar`

        :param once: Run the Field function once
        :type once: :class:`bool`
        '''
        thread = threading.Thread(
            target=self.run_threaded,
            args=(bar, run_once),
            name=self.name  # #NOTE Eventually, id(self)
        )
        return thread


type FieldPrecursor = FieldName | Field | FormatterFieldSig
from . import _types
FieldPrecursor = FieldPrecursor




'''Utility functions'''

#TODO Examples!

from copy import deepcopy
from ._types import PythonData

from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import Any


def join_options(
    it: Iterable[object],
    /,
    sep: str = ', ',
    final_sep: str = 'or',
    quote: bool = True,
    oxford: bool = True,
    limit: int = None,
    overflow: str = '...',
    metasep: str = ' '
) -> str:
    '''
    Tie together a list of objects for use in natural sentences.
    Commonly used to present valid variable options or expected values
    in error messages.

    :param it: The iterable of objects to join
    :type it: :class:`Iterable[object]`

    :param sep: The string separating every option,
        defaults to ``', '``
    :type sep: :class:`str`

    :param final_sep: The string separating the last two options,
        defaults to ``'or'``
    :type final_sep: :class:`str`

    :param quote: Put each option in quotes,
        defaults to ``True``
    :type quote: :class:`bool`

    :param oxford: Put `sep` between second-last option and `final_sep`,
        otherwise `metasep`, defualts to ``True``
    :type oxford: :class:`bool`

    :param limit: Show this many options before appending `overflow`,
        defaults to ``None``
    :type limit: :class:`int`

    :param overflow: The string appended when `limit` options are joined,
        defaults to ``'...'``
    :type sep: :class:`str`

    :param metasep: Separates `final_sep` and the last option
    :type metasep: :class:`str`
    '''
    if not hasattr(it, '__iter__'):
        raise TypeError(f"Can only join an iterable, not {type(it)}.")
    match tuple(it):
        case ():
            return ""
        case (only,):
            return repr(str(only)) if quote else str(only)
    opts = [repr(str(item)) if quote else str(item) for item in it][:limit]
    if limit is not None and len(opts) >= limit:
        opts.append(overflow)
    elif oxford:
        opts[-1] = metasep.join((final_sep, opts[-1]))
    else:
        opts[-1] = metasep.join((opts.pop(-2), final_sep, opts[-1]))
    return sep.join(opts)


def str_to_bool(value: str | bool, /) -> bool:
    '''Returns `True` or `False` bools for truthy or falsy strings.'''
    if isinstance(value, bool):
        return value
    value = str(value)
    truthy = "true t yes y on 1".split()
    falsy = "false f no n off 0".split()
    pattern = value.lower()
    if pattern not in truthy + falsy:
        raise ValueError(f"Invalid argument: {value!r}")
    return (pattern in truthy or not pattern in falsy)


def recursive_scrub(
    obj: Iterable,
    /,
    test: Callable[[Any], bool],
    inplace: bool = False,
) -> Iterable:
    '''Scrub an iterable of any elements that pass a callable predicate.

    By default, return a scrubbed copy of the original object.
    For dicts, remove whole items whose keys pass the predicate.
    Remove elements recursively.
    '''
    new = obj
    if not inplace:
        new = deepcopy(obj)

    def clean(o):
        if isinstance(o, list):
            i = 0
            while i < len(o):
                elem = o[i]
                if test(elem):
                    del o[i]
                    continue
                else:
                    clean(elem)
                i += 1

        elif isinstance(o, dict):
            for key, val in tuple(o.items()):
                if test(key):
                    del o[key]
                # elif test(val):
                    # del o[key]
                else:
                    clean(val)

        elif test(o):
            del o

    clean(new)
    return new


def scrub_comments(
    obj: Iterable,
    /,
    pattern: str | tuple[str] = '//',
    inplace: bool = False
) -> Iterable:
    '''Scrub an iterable of any elements that begin with a substring.

    By default, return a scrubbed copy of the original object.
    For dicts, remove whole items whose keys match the pattern.
    Remove elements recursively.
    '''
    predicate = (lambda o: True if (
        isinstance(o, str) and o.startswith(pattern)
        ) else False
    )
    return recursive_scrub(obj, test=predicate, inplace=inplace)


def make_error_message(
    cls: Exception,
    doing_what: str = None,
    blame: Any = None,
    expected: str = None,
    details: Iterable[str] = None,
    epilogue: str = None,
    file: str = None,
    line: int = None,
    indent: str = "  ",
    indent_level: int = 0
) -> Exception:
    '''Dynamically build an error message from various bits of context.

    Return an exception with the message passed as args.
    '''
    level = indent_level

    message = []
    if file is not None:
        message.append(f"In file {file!r}")
        if line is not None:
            message[-1] += f" (line {line})"
        message[-1] += ":"
        level += 1

    if line is not None:
        message.append(f"(line {line}):")
        level += 1

    if doing_what is not None:
        message.append(f"{indent * level}While {doing_what}:")

    level += 1

    if blame is not None:
        if expected is not None:
            message.append(
                f"{indent * level}Expected {expected}, "
                f"but got {blame} instead."
            )
        else:
            message.append(f"{indent * level}{blame}")

    if details is not None:
        message.append(
            '\n'.join(
                (indent * level + det)
                for det in details
            )
        )
        # message.append(
            # ('\n' + indent * level).join(details)
        # )

    if epilogue is not None:
        # message.append(level * indent + epilogue)
        message.append(epilogue)

    err = '\n' + ('\n').join(message)
    return cls(err)


def process_nested_dict(
    dct: dict | Any,
    roots: Sequence[str] = [],
    descended: int = 0
) -> tuple[list[list[str]], list[PythonData]]:
    '''
    Unravel nested dicts into keys and assignment values.

    :param dct: The nested dict to process, or an inner value of it
    :type dct: :class:`Mapping` | :class:`Any`

    :param roots: The keys which surround the current `dct`
    :type roots: :class:`Sequence`[:class:`str`], defaults to ``[]``

    :param descended: How far `dct` is from the outermost key
    :type descended: :class:`int`, defaults to ``0``

    :returns: Attribute names and assignment values
    '''
    nodes = []
    vals = []
    if not isinstance(dct, Mapping):
        # An assignment.
        return ([roots], [dct])

    if len(dct) == 1:
        # An attribute.
        for a, v in dct.items():
            roots.append(a)
            result = process_nested_dict(v, roots, descended)
            return result

    descended = 0  # Start of a tree.
    for attr, v in dct.items():
        roots.append(attr)
        if isinstance(v, Mapping):
            if descended < len(roots):
                descended = -len(roots)
            # Descend into lower tree.
            inner = process_nested_dict(v, roots, descended)
            inner_nodes, inner_vals = inner
            nodes.extend(inner_nodes)
            vals.extend(inner_vals)
        else:
            nodes.append(roots)
            vals.append(v)
        roots = roots[:-descended - 1]  # Reached end of tree.
    return nodes, vals


def nested_update(
    orig: Mapping | Any,
    upd: Mapping,
) -> dict:
    '''
    Merge two dicts and properly give them their innermost values.

    :param orig: The original dict
    :type orig: :class:`dict` | :class:`Any`

    :param upd: A dict that updates `orig`
    :type upd: :class:`dict`
    '''
    if not isinstance(orig, Mapping):
        # Replace the old value with the new:
        return upd

    for k, v in upd.items():
        if isinstance(v, Mapping):
            v = nested_update(orig.get(k, {}), v)
        orig[k] = v
    return orig




__all__ = (
    'FieldSpec',
    'BarSpec',
    'BarConfigSpec',
    'ConfigFileSpec',
    '_CmdOptionSpec',
)


from ast import Dict, List
from os import PathLike

from ._types import (
    Args,
    ASCII_Icon,
    ASCII_Separator,
    Field,
    FieldName,
    FormatStr,
    Icon,
    Kwargs,
    Separator,
    Unicode_Icon,
    Unicode_Separator,
)

from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import ParamSpec, Required, TypedDict, TypeVar


Bar = TypeVar('Bar')
P = ParamSpec('P')


class FieldConfigSpec(TypedDict, total=False):
    '''
    A dict specifying the structure of :class:`mybar.Field` constructor
    parameters as represented in config files and in the command line.
    '''
    name: FieldName
    icon: Icon | Sequence[ASCII_Icon, Unicode_Icon]
    template: FormatStr
    interval: float
    clock_align: bool
    overrides_refresh: bool
    threaded: bool
    always_show_icon: bool
    run_once: bool
    constant_output: str
    args: Args
    kwargs: Kwargs
    command: str
    script: PathLike
    allow_multiline: bool

class FieldSpec(FieldConfigSpec, total=False):
    '''
    A dict specifying the structure of :class:`mybar.Field` constructor
    parameters.
    '''
    func: Callable[P, str]
    setup: Callable[P, P.kwargs]
    bar: Bar

    unserializable = ('func', 'setup', 'bar')


class BarSpec(TypedDict, total=False):
    '''
    A dict specifying the structure of :class:`mybar.Bar` constructor
    parameters.
    '''
    break_lines: bool
    clock_align: bool
    count: int
    debug: bool
    join_empty_fields: bool
    override_cooldown: float
    refresh: float
    run_once: bool
    thread_cooldown: float
    unicode: bool

    '''
    The following field params are mutually exclusive with `template`:
    '''
    fields: Iterable[Field | FieldName]
    field_order: Sequence[FieldName]
    separator: Separator | Sequence[ASCII_Separator, Unicode_Separator]

    '''
    The `template` param is mutually exclusive with all field params:
    '''
    template: FormatStr


class BarConfigSpec(BarSpec, total=False):
    '''
    Specify options for :class:`mybar.bar.BarConfig` and config files.
    '''
    debug: bool
    field_definitions: Mapping[FieldName, FieldSpec]
    field_icons: Mapping[FieldName, Sequence[ASCII_Icon, Unicode_Icon] | Icon]
    field_order: Iterable[FieldName]
    from_icons: Mapping[FieldName, Icon]


class _CmdOptionSpec(TypedDict, total=False):
    '''
    Specify special command line optional arguments.
    '''
    config_file: PathLike
    dump_config: bool | int = 4




__all__ = (
    'FormatterFieldSig',
    'FmtStrStructure',
    'ConditionalFormatStr',
    'ElapsedTime',
    'format_uptime',
)


from string import Formatter

from .utils import join_options, make_error_message
from ._types import (
    Duration,
    FormatStr,
    FormatterLiteral,
    FormatterFname,
    FormatterFormatSpec,
    FormatterConversion,
    Duration
)

from collections.abc import Callable, Container, Hashable, Iterable, Iterator
from typing import Any, NamedTuple, NoReturn, Self, TypeAlias


class FormatStringError(ValueError):
    '''
    Base class for exceptions involving format strings.
    '''
    pass

class BrokenFormatStringError(FormatStringError):
    '''
    Raised when a format string cannot be properly parsed or contains
    positional fields (``'{}'``).
    '''
    pass

class InvalidFormatStringFieldNameError(FormatStringError):
    '''
    Raised when a format string has a field name not allowed or
    not defined by kwargs in a :meth:`str.format` call.
    '''
    pass

class InvalidFormatStringFormatSpecError(FormatStringError):
    '''
    Raised when a format string has a field with a format spec
    incompatible with string values in a :meth:`str.format` call.
    '''
    pass


### Format String Wonderland ###

class FormatterFieldSig(NamedTuple):
    '''
    A NamedTuple representing a format replacement field as generated
    by Formatter().parse().

    :param lit: Literal text preceding a replacement field
    :type lit: :class:`FormatterLiteral`

    :param name: The name of such a field, found inside curly braces
    :type name: :class:`FormatterFname`

    :param spec: The field's format spec, found inside curly braces
    :type spec: :class:`FormatterFormatSpec`

    :param conv: The field's conversion value, found inside curly braces
    :type conv: :class:`FormatterConversion`, optional
    '''

    lit: FormatterLiteral
    name: FormatterFname
    spec: FormatterFormatSpec
    conv: FormatterConversion

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.as_string()!r})"

    @classmethod
    def from_str(cls, fmt: FormatStr) -> Self:
        '''
        Convert a replacement field to a tuple of its elements.
        If there are multiple fields in the format string,
        only process the first one.

        :param fmt: The format string to convert
        :type fmt: :class:`FormatStr`
        '''
        try:
            parsed = tuple(Formatter().parse(fmt))

        except ValueError:
            err = f"Invalid format string: {fmt!r}"
            raise BrokenFormatStringError(err) from None

        if not parsed:
            err = f"The format string {fmt!r} contains no fields."
            raise FormatStringError(err)

        field = parsed[0]

        # Does the field have a fieldname?
        if field[1] == '':
            # No; it's positional.
            start = len(field[0])
            err = (
                f"The format string field at character {start} in {fmt!r} is "
                f"missing a fieldname.\n"
                 "Positional fields ('{}' for example) are not allowed "
                 "for this operation."
            )
            raise MissingFieldnameError(err)

        sig = cls(*field)
        return sig

    def as_string(
        self,
        with_literal: bool = True,
        with_conv: bool = True,
        with_spec: bool = True
    ) -> FormatStr:
        '''
        Recreate a format string field from a single field signature.

        :param with_literal: Include the signature's :class:`FormatterLiteral`,
            defaults to ``True``
        :type with_literal: :class:`bool`

        :param with_conv: Include the signature's :class:`FormatterConversion`,
            defaults to ``True``
        :type with_conv: :class:`bool`

        :returns: The format string represented by the signature
        :rtype: :class:`FormatStr`
        '''
        if self.name is None:
            # This format string has no fields whatsoever.
            # But it may have a literal:
            if with_literal:
                return self.lit
            return ''
        # Otherwise, fields that exist but are empty have a name of ''.

        inside_braces = self.name
        if with_conv and self.conv:
            inside_braces += '!' + self.conv
        if with_spec and self.spec:
            inside_braces += ':' + self.spec
        fmt = '{' + inside_braces + '}'

        return self.lit + fmt if with_literal else fmt


class FmtStrStructure(tuple[FormatterFieldSig]):
    '''
    Represents the structure of a whole format string broken up by
    :meth:`string.Formatter.parse` into a tuple of
    :class:`FormatterFieldSig`, one for each replacement field.
    '''
    def __repr__(self) -> str:
        return type(self).__name__ + tuple.__repr__(self)

    def get_names(self) -> tuple[FormatterFname]:
        '''
        Return a tuple with all the format string's field names in order.
        A positional field will have a name of ``''``.
        '''
        names = (name for sig in self if (name := sig.name) is not None)
        # names = (name for sig in self if (name := sig.name))
        return tuple(names)

    @classmethod
    def from_str(cls, template: FormatStr) -> Self:
        try:
            sigs = tuple(
                FormatterFieldSig(*field)
                for field in Formatter().parse(template)
            )
        except ValueError:
            err = f"Invalid format string: {template!r}"
            raise BrokenFormatStringError(err) from None

        return cls(sigs)

    def validate_fields(
        self,
        valid_names: Container[FormatterFname] | None = None,
        check_positionals: bool = False,
        check_specs: bool = False,
    ) -> bool | NoReturn:
        '''
        Ensure the represented format string has valid replacement fields.

        :param valid_names: Provide a collection of valid names to ensure
            that each field name refers to a valid :class:`Field` object,
            defaults to None.
            Raise :exc:`InvalidFormatStringFieldNameError` for any fields
            with names not in `valid_names`.
            When `valid_names` is ``None``, field names are not checked.
        :type valid_names: Container[:class:`FormatterFname`]

        :param check_positionals: Ensure each field has a name,
            defaults to ``False``.
            Raise :exc:`MissingFieldnameError`
            if any field is missing a field name.
        :type check_positionals: :class:`bool`

        :param check_specs: Ensure the format spec in each field is valid,
            defaults to ``False``.
            Raise :exc:`InvalidFormatStringFormatSpecError` if any field has an
            invalid format spec.
        :type check_specs: :class:`bool`

        :returns: ``True`` if the format string passes these checks,
            and raises exceptions if not.
        :rtype: :class:`bool` | :class:`NoReturn`

        :raises: :exc:`errors.MissingFieldnameError` when `template`
            contains positional fields
        :raises: :exc:`InvalidFormatStringFieldNameError`
            When `valid_names` is ``None``, field names are not checked.
        :raises: :exc:`errors.InvalidFormatStringFormatSpecError` when
            `template` contains fields with invalid format specs
        '''
        if not (check_positionals or check_specs) or valid_names is None:
            return True

        if check_positionals:
            if any(sig.name == '' for sig in self):
                # There's a positional field somewhere.
                # Issue an actionable error message:
                epilogue = (
                    "Bar format string fields must all have fieldnames."
                    "\nPositional fields ('{}' for example) are invalid."
                )
                raise MissingFieldnameError.with_highlighting(self, epilogue)

        # Test that all format fields do their formatting correctly:
        for sig in self:
            # Skip sigs that only have a literal:
            if sig.name is None:
                continue

            if valid_names is not None:
                if sig.name not in valid_names:
                    err = make_error_message(
                        InvalidFormatStringFieldNameError,
                        blame=repr(sig.name),
                        expected=f"one of {join_options(valid_names)}",
                        details=(
                            f"{sig.name!r} is not an allowed"
                            f" format field name."
                            ,
                        ),
                    )
                    raise err from None

            if check_specs:

# This isn't testable in general, only for Fields...
##                # Ideally, check the function to which each field refers:
##                field = valid_names.get(sig.name, {})
##                func = field.get('func', None)
##                if func is None:
##                    rtype = 'unknown'
##                else:
##                    rtype = func.__annotations__['return']

                try:
                    # Just assume that the value will be a string:
                    format(str(), sig.spec)
                except ValueError:
                    invalid_spec_msg = (
                        f"{sig.spec!r} is not a valid format spec"
                        f" when formatting strings"
                        ,
                        f"with format field {sig.as_string()!r}."
                    )
                    err = make_error_message(
                        InvalidFormatStringFormatSpecError,
                        details=invalid_spec_msg,
                    )
                    raise err from None

        return True


class MissingFieldnameError(FormatStringError):
    '''
    Raised when a format string field lacks a fieldname (i.e. is positional)
    when one is expected.

    '''
    @classmethod
    def with_highlighting(
        cls,
        sigs: FmtStrStructure,
        epilogue: str = ''
    ) -> Self:
        '''
        Make an error message with highlighting under positional fields.

        :param sigs: The problematic format string in
            the form of a :class:`FmtStrStructure`
        :type sigs: :class:`FmtStrStructure`

        :param epilogue: A message at the end of the traceback
        :type epilogue: :class:`str`

        :returns: A new :class:`errors.MissingFieldnameError`
        :rtype: :class:`errors.MissingFieldnameError`
        '''
        rebuilt = ""
        highlight = " "  # Account for the repr quotation mark.

        for sig in sigs:
            field = sig.as_string()
            rebuilt += field

            if sig.name == '':  # For each positional field...
                # Skip over the part not in braces:
                highlight += " " * len(sig.lit)

                # Only highlight the part in braces.
                bad_field_len = len(field) - len(sig.lit)
                highlight += "^" * bad_field_len

            else:
                # Skip the whole field.
                highlight += " " * len(field)

        err = '\n'.join((
            "",
            "The following fields are missing fieldnames:",
            repr(rebuilt),
            highlight,
            epilogue
        ))

        return cls(err)


class ConditionalFormatStr:
    '''
    Reinterpret format strings based on the data they reference.
    Values in a mapping which are predicated ``False`` have their
    groupings shown blank when the mapping is formatted.

    :param fmt: The initial format string
    :type fmt: :class:`FormatStr`

    :param sep: Surrounds related fields and literal text that should be
        grouped together, defaults to ``":"``
    '''
    def __init__(self,
        fmt: FormatStr,
        sep: str = ':'  # Other common values are , /
    ) -> None:
        self.fmt = fmt
        self.sep = sep
        self.fnames, self.groups = self.parse()

    def parse(self,
        sep: str = None
    ) -> tuple[tuple[FormatterFname], FmtStrStructure]:
        '''
        Parse a format string using its format fields and a separator.

        :param sep: Surrounds related fields and literal text that should be
            grouped together, defaults to ``self.sep``
        :type sep: :class:`str`, optional

        :returns: A nested tuple of the string's fieldnames
            and its :class:`FmtStrStructure`
        :rtype: :class:`tuple[tuple[FormatterFname], FmtStrStructure]`
        '''
        if sep is None:
            sep = self.sep

        sections = []
        # Split fmt for parsing, but join any format specs that get broken:
        pieces = (p for p in self.fmt.split(sep))

        def _is_malformed(piece: FormatStr):
            '''Return whether a format string is malformed.'''
            try:
                tuple(Formatter().parse(piece))
            except ValueError:
                return True
            else:
                return False

        try:
            for piece in pieces:
                while _is_malformed(piece):
                    # Raise StopIteration if a valid field end is not found:
                    piece = sep.join((piece, next(pieces)))
                sections.append(piece)

        except StopIteration:
            exc = make_error_message(
                BrokenFormatStringError,
                doing_what=f"parsing {self!r} format string {self.fmt!r}",
                details=[
                    f"Invalid fmt substring begins near ->{piece!r}"
                ]
            )
            raise exc from None

        groups = tuple(
            tuple(Formatter().parse(section))
            for section in sections
        )

        fnames = tuple(
            name
            for section in groups
            for parsed in section
            if (name := parsed[1])
        )

        return fnames, groups

    def format(self,
        namespace: dict[FormatterFname, Any],
        predicate: Callable[[Any], bool] = bool,
        sep: str = None,
        substitute: str = None,
        round_by_default: bool = True,
    ) -> str:
        '''Format a dict of numbers according to a format string by parsing
        fields delineated by a separator `sep`.
        Field groups which fail the `predicate` are not shown in the
        final output string.
        If specified, `substitute` will replace invalid fields instead.

        :param namespace: A mapping with values to which fieldnames refer
        :type namespace: :class:`dict[FormatterFname]`

        :param predicate: A callable to determine whether a field should be
            printed, defaults to :func:`bool`
        :type predicate: :class:`Callable[[Any], bool]`

        :param sep: Used to join field groups into the final string,
            defaults to ``self.sep``
        :type sep: :class:`str`

        :param substitute: When set, replaces a field group that fails `predicate`
        :type substitute: :class:`str`, optional

        :param round_by_default: Round values if they are floats,
            defaults to ``True``
        :type round_by_default: :class:`bool`

        :returns: A string with field groups hidden which don't pass `predicate`
        :rtype: :class:`str`
        '''
        if sep is None:
            sep = self.sep

        newgroups = []
        for i, group in enumerate(self.deconstructed):
            if not group:
                # Just an extraneous separator.
                newgroups.append(())
                continue

            newgroup = []

            for maybe_field in group:
                # Handle sections that do not pass the predicate:
                if not predicate(val := namespace[maybe_field[1]]):

                    ##
                    # Maybe check the length to see if a regular .format() can be used!
                    ##

                    if substitute is not None:
                        newgroups.append(substitute)
                    break

                buf = ""

                match maybe_field:
                    case [lit, None, None, None]:
                        # A trailing literal.
                        # Only append if the previous field was valid:
                        buf += lit

                    case [lit, field, spec, conv]:
                        # A veritable format string field!
                        # Add the text right before the field:
                        if lit is not None:
                            buf += lit

                        # Format the value if necessary:
                        if spec:
                            buf += format(val, spec)
                        elif round_by_default and isinstance(val, float):
                            buf += str(round(val))
                        else:
                            buf += str(val)

                    case weird:
                        raise ValueError(
                            f"\n"
                            f"Invalid structure in tuple\n"
                            f"  {i} {maybe_field}:\n"
                            f"  {weird!r}"
                        )

                if buf:
                    newgroup.append(buf)
            if newgroup:
                newgroups.append(newgroup)

        # Join everything.
        return sep.join(''.join(g) for g in newgroups)


class ElapsedTime:
    '''
    Represent elapsed time in seconds with a variety of larger units.
    '''
    YEARS = 12*4*7*24*60*60
    MONTHS = 4*7*24*60*60
    WEEKS = 7*24*60*60
    DAYS = 24*60*60
    HOURS = 60*60
    MINS = 60
    SECS = 1

    conversions_to_secs = {
        'years': YEARS,
        'months': MONTHS,
        'weeks': WEEKS,
        'days': DAYS,
        'hours': HOURS,
        'mins': MINS,
        'secs': SECS,
        'femtofortnights': 2 * WEEKS * 10**-15  # You can't see this.
    }
    '''
    A mapping of units of elapsed time to their equivalents in seconds.
    '''
    class dict(dict):
        '''
        Hide the silly unit.
        '''
        class SpecialStr(str):
            def __lt__(self, other) -> bool:
                '''
                Intercept sorted() (used by Sphinx).
                '''
                return False

        def __init__(self, *args, **kwargs) -> None:
            if args:
                args = ({self.SpecialStr(k): v for k, v in args[0].items()},)
            super().__init__(*args, **kwargs)

        def __iter__(self) -> Iterator:
            '''
            Intercept Sphinx.
            '''
            import inspect
            frame = inspect.getouterframes(inspect.currentframe())[1]
            if frame.function == 'object_description':
                self.pop('femtofortnights', None)  # Shhh, it's a secret.
            return super().__iter__()

        def __repr__(self) -> str:
            return repr(self._safe())

        def _safe(self) -> Self:
            hide_meme = self.copy()
            hide_meme.pop('femtofortnights', None)  # Shhh, it's a secret.
            return hide_meme

    conversions_to_secs = dict(conversions_to_secs)

    @classmethod
    def in_desired_units(cls,
        secs: int,
        units: tuple[Duration]
    ) -> dict[Duration, int]:
        '''
        Convert seconds to multiple units of time.

        The resulting value of the smallest unit in `units` is kept as
        a :class:`float`.
        Smaller units overflow into larger units when they meet or exceed
        the threshold given by
        :obj:`ElapsedTime.conversions_to_secs[larger_unit]`,
        but only if a larger unit is present in `units`.

        For example...

        Running ``in_desired_units(12345, ('mins', 'hours'))``
             yields ``{'hours': 3, 'mins': 25.75}``,

        but ``in_desired_units(12345, ('mins', 'days'))``
             yields ``{'days': 0, 'mins': 205.75}``,

        and ``in_desired_units(12345, ('hours',))``
             yields ``{'hours': 3.4291666666666667}``.

        :param secs: Seconds to be converted
        :type secs: :class:`int`

        :param units: Units to convert
        :type units: :class:`tuple[Duration]`

        :returns: A mapping of unit names to amount of time in those units
        :rtype: :class:`dict[Duration, int]`
        '''
        if not all(u in cls.conversions_to_secs for u in units):
            # At least one unit wasn't recognized. Raise an error:
            valid = cls.conversions_to_secs._safe()
            exptd = join_options(valid)
            unrec = join_options(set(units) - set(valid))
            exc = make_error_message(
                LookupError,
                blame=repr(units),
                expected=f"a sequence of unit names from {exptd}",
                details=[
                    f"The following unit names are not recognized:",
                    f"{unrec}",
                ]
            )
            raise exc

        # Get the units in order of largest first:
        ordered = tuple(u for u in cls.conversions_to_secs if u in units)

        table = {}
        if len(ordered) == 1:
            unit = ordered[0]
            # Avoid robbing the only unit of its precision. Just divide:
            table[unit] = secs / cls.conversions_to_secs[unit]
            return table

        for unit in ordered[:-1]:
            table[unit], secs = divmod(secs, cls.conversions_to_secs[unit])
        # Give the least significant unit a precise value:
        last_u = ordered[-1]
        table[last_u] = secs / cls.conversions_to_secs[last_u]
        return table


def format_uptime(
    secs: int,
    sep: str,
    namespace: dict[Duration, int],
    groups: tuple[tuple[FormatterFieldSig]],
    *args,
    **kwargs
) -> str:
    '''
    Format a dict of numbers according to a format string by parsing
    fields delineated by a separator.

    :param secs: Total elapsed time in seconds (unused)
    :type secs: :class:`int`

    :param sep: A string that separates groups of text based on division
        of time
    :type sep: :class:`str`

    :param namespace: A mapping of time unit names to :class:`int`
    :type namespace: dict[:class:`Unit`, :class:`int`]

    :param groups: A format string broken up by
        :func:`_setups.setup_uptime` into tuples of
        :class:`FormatterFieldSig` based on the locations of `separator`
    :type groups: tuple[tuple[:class:`FormatterFieldSig`]]
    '''
    newgroups = []
    for i, group in enumerate(groups):
        if not group:
            # Just an extraneous separator.
            newgroups.append(())
            continue

        newgroup = []

        for maybe_field in group:
            # Skip groups that should appear blank:
            if (val := namespace.get(maybe_field[1])) is not None and val < 1:
                break

            buf = ""

            match maybe_field:
                case [lit, None, None, None]:
                    # A trailing literal.
                    buf += lit

                case [lit, field, spec, conv]:
                    # A veritable format string field!
                    # Add the text right before the field:
                    if lit is not None:
                        buf += lit

                    # Format the value if necessary:
                    if spec:
                        buf += format(val, spec)
                    else:
                        try:
                            # Round down by default:
                            buf += str(int(val))
                        except TypeError:
                            buf += str(val)

                case _:
                    raise ValueError(
                        f"\n"
                        f"Invalid structure in tuple\n"
                        f"  {i} {maybe_field}:\n"
                        f"  {spam!r}"
                    )

            if buf:
                newgroup.append(buf)
        if newgroup:
            newgroups.append(newgroup)

    # Join everything.
    return sep.join(''.join(g) for g in newgroups)


##from typing import NamedTuple
### class Context(NamedTuple):
##    # value: str
##class Context(dict):
##    def __init__(
##        self,
##        *args,
##        value: str = None,
##    ) -> None: 
##        super().__init__(self)
##        self.value = value
##        # self.
##
##class IconFactory:
##    def __init__(
##        self,
##        func: Callable[[Context], Icon],
##        default: Icon = "",
##    ) -> None:
##        self.func = func
##        self.default = default
##
##    def interpret(self, ctx: Context = None, default: Icon = None) -> Icon:
##        if default is None:
##            default = self.default
##        if ctx is None:
##            return default
##        try:
##            return self.func(ctx)
##        except Exception:
##            return default
##
##
### class Icon:
### class IconMatcher(dict):
### class IconDict(dict):
##class EasyIcon(dict):
##
##    def __init__(
##        self,
##        map: dict[Hashable, Icon] = {},
##        default: Icon = "",
##        # picker = None
##    ) -> None:
##        # self._registry = statemap
##        if statemap:
##            self.update(statemap)
##        self.default = default
##        # if callable(picker):
##            # self.picker = picker
##        # return self
##
####    def __setitem__(self, state: Any, var: str) -> None:
####        self._registry[id(state)] = var
##        # self[id(state)] = var
##
####    def __getitem__(self, state: Any, ) -> str:
####        return self._registry.get(state)
##    def choose(self, obj: Any):
##        return self.get(obj, self.default)
##
##    def interpret(self, key: Any):
##        return self.get(self.picker(key), self.default)
##
##
##    def __repr__(self) -> str:
##        # return self._registry[self._default]
##        stuff = ', '.join(' on '.join((repr(v), repr(k))) for k, v in self.items())
##        cls = type(self).__name__
##        return f"{cls}({stuff}, default={self.default!r})"
##
##    # def __str__(self) -> str:
##        # return self._registry[self._default]
##
####cb = Callable[[Kwargs], State1|State2]
####result, state = cb()
####form = self.fmt.interpret(result, state)
####form = self.format(result, state)
##
##
##class Format:
##    content_key: str = '$'
##    def __init__(
##        self,
##        # statemap: dict = {},
##        icon="",
##        default_fmt: str = "{icon}{"+content_key+'}',
##        fallback: str = "",
##        # custom_does_it_all: Callable[['result', 'context'], str] = None
##    ) -> None:
##        # self.statemap = statemap
##        self.icon = icon
##        self.default_fmt = default_fmt
##        self.fallback = fallback
##
##
### class 
##class FormatSwitcher(dict):
### class FormatSwitcher(Format):
##    content_key: str = '$'
##    def __init__(self,
##        statemap: dict['result', str] = {},
##        default: str = "{"+content_key+"}",
##        only_switch_icons: bool = True,
##        fallback: str = "",
##    ) -> None:
##        # self.statemap = statemap
##        self.icon = icon
##        self.default = default
##        self.fallback = fallback
##        self.update(statemap)
##
##    def format(self, result) -> str:
##        if isinstance(result, str):
##            return self.get(result, self.defaul)
##
##        icon, fmt = self.statemap.get(context, (self.icon, self.default_fmt))
##        contents = fmt.format_map({self.content_key: result, 'icon': icon})
##        return contents
##
##
##class FormatMaker(Format):
##    def __init__(self,
##        fallback: str = "",
##        custom_does_it_all: Callable[['result'], str] = None
##    ) -> None:  
##        super().__init__(icon, default_fmt, fallback)
##
##        # if custom_does_it_all is None:
##            # custom_does_it_all = self._default_thingy
##
##    @staticmethod
##    def _default_thingy(fmt, result, context) -> str:
##        contents = fmt.format_map({content_key: result, 'icon': icon})
##        return
##
##    def format(self, result, context):
##        icon, fmt = self.statemap.get(context, (self.icon, self.default_fmt))
##        contents = fmt.format_map({self.content_key: result, 'icon': icon})
##        return contents
##
##def check_battery(ctx):
##    chrg_icn = ""
##    if ctx.get('charging'):
##        return chrg_icn + " "
##
##    icon_bank = """
##    
##    
##    
##    
##    
##    """.split()
##    def mapper(n):
##        icon = ""
##        for i, test in enumerate((10, 25, 50, 75, 100)):
##            if n <= test:
##                icon = icon_bank[i]
##                return icon + " "





__all__ = (
    'BrokenFormatStringError',
    'CLIFatalError',
    'CLIUsageError',
    'CompatibilityWarning',
    'DefaultFieldNotFoundError',
    'FailedSetup',
    'FormatStringError',
    'IncompatibleArgsError',
    'InvalidArgError',
    'InvalidBarError',
    'InvalidFieldError',
    'InvalidFieldSpecError',
    'InvalidFormatStringFieldNameError',
    'InvalidOutputStreamError',
    'MissingFieldnameError',
    'UndefinedFieldError',
)


from os import PathLike
from typing import Any, IO
from .formatting import (
    FormatStringError,
    BrokenFormatStringError,
    InvalidFormatStringFieldNameError,
    InvalidFormatStringFormatSpecError,
    MissingFieldnameError
)
from ._types import Contents


class DefaultFieldNotFoundError(LookupError):
    '''
    Raised for references to an undefined default :class:`Field`.
    '''
    pass

class IncompatibleArgsError(ValueError):
    '''
    Raised when a class is instantiated with one or more missing or
    incompatible parameters.
    '''
    pass

class InvalidArgError(ValueError):
    '''
    Raised by a field function when it receives an invalid argument.
    '''
    pass

class InvalidFieldError(TypeError):
    '''
    Raised when a field is either not an instance of :class:`Field` or a string not
    found in the default fields container.
    '''
    pass

class InvalidFieldSpecError(TypeError):
    '''
    Raised when an expected :class:`FieldSpec` has the wrong type or an invalid structure.
    '''
    pass

class InvalidBarError(AttributeError):
    '''
    Raised when Field._check_bar() finds missing attributes in a
    potential status bar.
    '''

class InvalidOutputStreamError(AttributeError):
    '''
    Raised when an :class:`IO` stream lacks ``write()``,
    ``flush()`` and ``isatty()`` methods.
    '''
    pass

class UndefinedFieldError(LookupError):
    '''
    Raised if, when parsing a config file, a field name appears in the
    `field_order` item of the :class:`dict` passed to :meth:`Bar.from_dict()` that is neither
    found in its `field_definitions` parameter nor in :attr:`Field._default_fields`.
    '''
    pass


class CLIFatalError(Exception):
    '''
    Base class for errors that cause the CLI program to exit.

    :param msg: The error message to issue when exiting
    :type msg: :class:`str`
    '''
    def __init__(self, msg: str) -> None:
        super().__init__()
        self.msg = msg

    def __str__(self):
        return self.msg


class CLIUsageError(CLIFatalError):
    '''
    Raised when the CLI program is used incorrectly.
    '''
    pass


class FailedSetup(Exception):
    '''
    Raised by a setup function when it cannot return a suitable value.
    :meth:`Field.run` uses `backup` for the :class:`Bar` buffer value instead.

    :param backup: The field contents to use instead of the setup function return value
    :type backup: :class:`str`
    '''
    def __init__(self, backup: Contents) -> None:
        self.backup = backup


class CompatibilityWarning(Warning):
    '''
    Raised when certain application features will break if used in an
    environment without proper support.
    '''



# import psutil

from .errors import (
    BrokenFormatStringError,
    FailedSetup,
    IncompatibleArgsError,
)
from .formatting import ConditionalFormatStr
from .utils import join_options, make_error_message 
from ._types import FormatStr


async def setup_uptime(
    fmt: FormatStr,
    sep: str = None,
    *args,
    **kwargs
) -> dict[str]:
    setupvars = {'fmt': fmt, 'sep': sep}
    try:
        conditional = ConditionalFormatStr(fmt, sep)
    except BrokenFormatStringError:
        raise FailedSetup(backup=fmt)
    else:
        setupvars.update({
            'fnames': conditional.fnames,
            'groups': conditional.groups,
        })
    return setupvars


def setup_mpris() -> dict[str]:
# async def setup_mpris(service: str | None = None) -> dict[str]:
    '''
    '''
    import pympris
    import dbus
    from dbus.mainloop.glib import DBusGMainLoop

    dbus_loop = DBusGMainLoop()
    bus = dbus.SessionBus(mainloop=dbus_loop)

    player_ids = list(pympris.available_players())
    type ServiceName = str
    players: dict[ServiceName] = {
        p.root.Identity: pympris.MediaPlayer(p, bus)
        for p in player_ids
    }

    setupvars = {
        'players': players,
    }
    return setupvars


##def setup_cpu_usage(
##    in_fahrenheit: bool = False,
##    *args,
##    **kwargs
##):
##    setupvars = {}
##
##    temps = psutil.sensors_temperatures(in_fahrenheit)
##    match temps:
##        case {'k10temp': t} | {'coretemp': t}:
##            current = t[0].current
##        case _:
##            current = '??'
##
##    return fmt.format(current, symbol)

##def setup_mem_usage(
##    prec: int = 1,
##    measure: str = 'used',
##    unit: str = 'G',
##    fmt: str = "{:.{}f}{}",
##    *args,
##    **kwargs
##):
##    setupvars = {}
##
##    if unit not in UNITS:
##        raise InvalidArgError(
##            f"Invalid unit: {unit!r}\n"
##            f"'unit' must be one of "
##            f"{join_options(UNITS, quote=True)}."
##        )
##
##    disk = psutil.disk_usage(path)
##    statistic = getattr(disk, measure, None)
##    if statistic is None:
##        raise InvalidArgError(
##            f"Invalid measure on this operating system: {measure!r}.\n"
##            f"measure must be one of "
##            f"{join_options(statistic._fields, quote=True)}"
##        )
##
##    setupvars.update(
##        unit=unit,
##        measure=measure,
##    )




__all__ = (

    'ASCII_Icon',
    'ASCII_Separator',
    'Args',
    'AssignmentOption',
    'Bar',
    'BarConfig',
    # 'BatteryStates',
    # 'ColorEscaping',
    'ConsoleControlCode',
    'Contents',
    # 'Context',
    'Duration',
    'Field',
    'FieldFuncSetup',
    'FieldFunc',
    'FieldName',
    'FieldName',
    'FieldOrder',
    'FieldPrecursor',
    'FileContents',
    'FormatStr',
    'FormatterFieldSig',
    'FormatterConversion',
    'FormatterFname',
    'FormatterFormatSpec',
    'FormatterLiteral',
    'HostOption',
    'Icon',
    'JSONText',
    'Kwargs',
    'Line',
    'MetricSymbol',
    'NmConnIDSpecifier',
    'NmConnFilterSpec',
    'OptName',
    'OptSpec',
    'Pattern',
    'PythonData',
    'Separator',
    'StorageMeasure',
    'Unicode_Icon',
    'Unicode_Separator',

)


from collections.abc import Callable, Mapping
from enum import Enum
from os import PathLike
from re import Pattern as re_Pattern
from typing import (
    Any,
    Literal,
    NamedTuple,
    Required,
)


POWERS_OF_1024 = {
    'K': 1024**1,
    'M': 1024**2,
    'G': 1024**3,
    'T': 1024**4,
    'P': 1024**5,
}


type ASCII_Icon = str
type ASCII_Separator = str
type Args = list
type AssignmentOption = re_Pattern[r'(?P<key>\w+)=(?P<val>.*)']
type Bar = 'Bar'
type BarConfig = 'BarConfig'
type ConsoleControlCode = str
type Contents = str

type Duration = Literal[
    'secs',
    'mins',
    'hours',
    'days',
    'weeks',
    'months',
    'years'
]
'''Possible names for units of elapsed time.'''

type Field = 'Field'
type FieldFuncSetup[**P] = Callable[P, P.kwargs]
type FieldFunc[**P] = Callable[P, str]
type FieldName = str
type FieldName = str
type FieldOrder = tuple[FieldName]
type FieldPrecursor = FieldName | Field | FormatterFieldSig
type FileContents = str
type FormatStr = str
type FormatterFieldSig = 'FormatterFieldSig'

type FormatterConversion = str | None
'''
The `conversion` part of one tuple in the iterable returned
by :func:`string.Formatter.parse()`.
'''

type FormatterFname = str | None
'''
The `field_name` part of one tuple in the iterable returned
by :func:`string.Formatter.parse()`.
'''

type FormatterFormatSpec = str | None
'''
The `format_spec` part of one tuple in the iterable returned
by :func:`string.Formatter.parse()`.
'''

type FormatterLiteral = str | None
'''
The `literal_text` part of one tuple in the iterable returned
by :func:`string.Formatter.parse()`.
'''

type HostOption = Literal[
    'nodename',
    'sysname',
    'release',
    'version',
    'machine',
]

type Icon = str
type JSONText = str
type Kwargs = dict[str]
type Line = str
type MetricSymbol = Literal[*POWERS_OF_1024.keys()]

type NmConnIDSpecifier = Literal['id', 'uuid', 'path', 'apath']
'''
One of several keywords NetworkManager provides to narrow down
connection results. From the ``nmcli`` man page:

.. code-block:: none

   id, uuid, path and apath keywords can be used if ID is
   ambiguous. Optional ID-specifying keywords are:

   id
       the ID denotes a connection name.

   uuid
       the ID denotes a connection UUID.

   path
       the ID denotes a D-Bus static connection path in the format of
       /org/freedesktop/NetworkManager/Settings/num or just num.

   apath
       the ID denotes a D-Bus active connection path in the format of
       /org/freedesktop/NetworkManager/ActiveConnection/num or just num.
'''

type NmConnFilterSpec = Mapping[NmConnIDSpecifier, str]
'''
A mapping passed to :func:`mybar.field_funcs.get_net_stats` to filter
multiple connections.
'''

type OptName = str
type OptSpec = dict[OptName, Any]
type Pattern = str
type PythonData = object
type Separator = str
type StorageMeasure = Literal['total', 'used', 'free', 'percent']
type Unicode_Icon = str
type Unicode_Separator = str


# Used in the future by field_funcs:
class Context(NamedTuple):
    '''
    For future use.
    '''
    contents: str = None
    state: Any = None


class BatteryStates(Enum):
    '''
    For future use.
    '''
    CHARGING = 'charging'
    DISCHARGING = 'discharging'
    # Progressive/dynamic battery icons!
        # 
        # 
        # 
        # 
        # 


class ColorEscaping(Enum):
    '''
    For future use.
    '''
    ANSI = 'ANSI'
    POLYBAR = 'POLYBAR'





import os.path
import sys

from .bar import Bar, BarConfig
from .cli import ArgParser, CLIUsageError
from .constants import CONFIG_FILE


def main() -> None:
    '''Run the command line utility.'''
    try:
        parser = ArgParser()
        try:
            bar_options, command_options = parser.parse_args()
        except CLIUsageError as e:
            parser.error(e.msg)  # Shows usage

        file = command_options.pop('config_file', CONFIG_FILE)
        absolute = os.path.abspath(os.path.expanduser(file))

        try:
            config = BarConfig.from_file(absolute, overrides=bar_options)

        # Handle missing config files:
        except FileNotFoundError as e:
            config = BarConfig(bar_options)
            if 'dump_config' not in command_options:
                # Skip write if we're just printing the config.
                config.write_with_approval(absolute)

        # Permissions error:
        except OSError as e:
            e.add_note("Exiting...")
            raise e from None

        except KeyboardInterrupt:
            parser.quit()

        # Handle options that alter the behavior of the command itself:
        if 'dump_config' in command_options:
            indent = command_options.pop('dump_config', None)
            parser.quit(BarConfig.as_json(config, indent=indent))

        bar = Bar.from_config(config)
        bar.run()

    except KeyboardInterrupt:
        print()
        sys.exit(1)


if __name__ == '__main__':
    main()




__all__ = (
    'CONFIG_FILE',
    'DEBUG',
    'CSI',
    'CLEAR_LINE',
    'HIDE_CURSOR',
    'UNHIDE_CURSOR',
)


import sys
import os.path

from ._types import ConsoleControlCode


CONFIG_FILE: str = (
    '~/.config/mybar/mybar.conf'
    # '~/.config/mybar/conf.json'
)
'''The default mybar config file path.'''

DEBUG: bool = False
'''The default debug state.'''


# Used by Bar:
CSI: ConsoleControlCode = '\033['
'''Unix terminal escape code (control sequence introducer).'''

CLEAR_LINE: ConsoleControlCode = '\x1b[2K'
'''VT100 escape code to clear the line.'''

HIDE_CURSOR: ConsoleControlCode = '?25l'
'''VT100 escape code to hide the cursor.'''

UNHIDE_CURSOR: ConsoleControlCode = '?25h'
'''VT100 escape code to unhide the cursor.'''




"""
######
mybar
######

*Craft highly customizable status bars with ease.*

**mybar** is a versatile status bar library and command line tool
written in Python.


:copyright: (c) 2021-present akyuute.
:license: MIT, see LICENSE for more details.
"""

__title__ = 'mybar'
__description__ = "Craft highly customizable status bars with ease."
__url__ = "https://github.com/akyuute/mybar"
__version__ = '0.14'
__author__ = "akyuute"
__license__ = 'MIT'
__copyright__ = "Copyright (c) 2021-present akyuute"


__all__ = (
    'Bar',
    'BarConfig',
    'CONFIG_FILE',
    'DEBUG',
    'Field',
    'run',
    'write_initial_config'
)


from os import PathLike

from .constants import CONFIG_FILE, DEBUG
from .bar import Bar, BarConfig
from .field import Field


def run(*, once: bool = False, file: PathLike = None) -> None:
    '''
    Generate a new :class:`Bar` from a config file and run it in STDOUT.
    Ask to write the file if it doesn't exist.

    :param once: Print the bar only once, defaults to ``False``
    :type once: :class:`bool`

    :param file: The config file to source,
        defaults to :obj:`CONFIG_FILE`
    :type file: :class:`PathLike`
    '''
##    import __main__ as possibly_repl
##    if not hasattr(possibly_repl, '__file__'):
##        # User is in a REPL
    try:

        try:
            bar = Bar.from_file(file)
        except FileNotFoundError as e:
            _p_ = __package__
            msg = (
                f"Try running `{_p_}.write_initial_config()`"
                f" if this is your first time using {_p_}."
            )
            e.add_note(msg)
            raise e from None

        bar.run(once=once)

    except KeyboardInterrupt:
        print()
        return


def write_initial_config() -> None:
    '''
    Write a new default config file after getting user approval.
    '''
    BarConfig.write_with_approval(CONFIG_FILE)


del PathLike




__all__ = (
    'ArgParser',
    'OptionsAsker',
)


import os.path
from argparse import ArgumentParser, SUPPRESS, Namespace, HelpFormatter
from enum import Enum
from os import PathLike

from . import __version__
from .constants import CONFIG_FILE
from .errors import CLIUsageError
from .namespaces import BarConfigSpec, _CmdOptionSpec, FieldConfigSpec
from .utils import str_to_bool
from ._types import (
    AssignmentOption,
    FieldName,
    OptName,
    OptSpec,
)

from typing import Any, Callable, Iterable, Never, TypeAliasType


PROG = __package__


class ArgFormatter:
    '''
    Methods for formatting args.
    '''
    @staticmethod
    def SplitFirst(char: str) -> Callable[[str], str]:
        '''
        Give this to the `type` parameter of
        :func:`ArgumentParser.add_argument` and it will split args using
        `char`.
        '''
        return (lambda f: f.split(char, 1))

    @staticmethod
    def ToTuple(length: int) -> Callable[[Any], tuple]:
        '''
        Give this to the `type` parameter of
        :func:`ArgumentParser.add_argument` and it will fill tuples with
        `length` copies of the arg.
        '''
        return (lambda s: (s,) * length)


class ArgParser(ArgumentParser):
    '''
    A custom command line parser used by the command line utility.
    '''
    assignment_arg_map = {
        'icon_pairs': 'field_icons',
        'from_icons': ('field_icons', 'field_order'),
    }

    def __init__(self) -> None:
        super().__init__(
            prog=PROG,
            argument_default=SUPPRESS,
        )
        self.add_arguments()

    def parse_args(
        self,
        args: list[str] = None
    ) -> tuple[BarConfigSpec, _CmdOptionSpec]:
        '''
        Parse command line arguments and return a dict of options.
        This will additionally process args with key-value pairs.
        If `args` is None, process from STDIN.

        :param args: The args to parse, get from STDIN by default
        :type args: :class:`list[str]`
        '''
        # Use vars() because dict items are more portable than attrs.
        opts = vars(super().parse_args(args))
        params = self.process_assignment_args(opts)

        # For options that control specific fields:
        if 'field_options' in params:
            fields = self.process_field_options(
                params.pop('field_options')
            )
            params['field_definitions'] = fields

        # For options that control what the command does:
        keys = (
            _CmdOptionSpec.__optional_keys__ | _CmdOptionSpec.__required_keys__
        )
        cmd_options = {k: params.pop(k) for k in keys if k in params}
        return params, cmd_options

    def process_assignment_args(
        self,
        opts: BarConfigSpec,
        dest_map: dict[OptName, OptName | tuple[OptName]] = None
    ) -> BarConfigSpec:
        '''
        Make dicts from key-value pairs in assignment args.
        Split strings on '=' and convert the resultant keys and values to dicts.

        ``['foo=bar', 'baz=grok']`` -> ``{'foo': 'bar', 'baz': 'grok'}``

        :param opts: The Bar options dict
        :type opts: :class:`BarConfigSpec`

        :param dest_map: A dict mapping `dest` names to names of new dict
            items in which to save key-value pairs, defaults
            to :attr:`Parser.assignment_arg_map`
        :type dest_map: dict[:class:`OptName`,
            :class:`OptName` | tuple[:class:`OptName`]]
        '''
        if dest_map is None:
            dest_map = self.assignment_arg_map

        for src, new_dest in dest_map.items():
            if (vals := opts.pop(src, None)):
                assignments = dict(
                    pair for item in vals
                    # Only use items that are pairs:
                    if len(pair := item.split('=', 1)) == 2
                )
                if isinstance(new_dest, str):
                    opts[new_dest] = assignments
                else:
                    for place in new_dest:
                        opts[place] = assignments
        return opts

    @staticmethod
    def process_field_options(
        options: Iterable[AssignmentOption]
    ) -> dict[FieldName, FieldConfigSpec]:
        '''
        Parse variable assignment args given for Field definitions.
        Each option in `options` should look like 'key=val'.
        We specifically handle Field options. To that end, parse nested
        options with emulated dotted attribute access by ignoring all
        options not of the form 'field.key=val'.
        After parsing options, return a dict mapping them to Field names,
        similar to `field_definitions` in :meth:`Bar.from_config()`.

        I think this has some logical chinks. Definitely overengineered.

        :param options: Assignment option 'key=val' pairs to parse
        :type options: Iterable[:class:`AssignmentOption`]

        :returns: a dict mapping Field names to dicts of options
        :rtype: dict[:class:`FieldName`, :class:`FieldConfigSpec`]
        '''
        field_definitions = {}
        for opt in options:
            match (pair := opt.split('=', 1)):

                case [positional]:  # Looks like '--opt'
                    # Skip command options.
                    continue

                case [field_and_opt, val]:  # Looks like 'key=val'
                    if field_and_opt.isidentifier():
                        # No dots means it's an option for the Bar.
                        # Not our problem.
                        continue

                    # Looks like 'field.key=val'
                    # An option for a Field.
                    # Handle attribute access through dots:
                    field_name, field_opt = field_and_opt.split('.', 1)
                    if field_name not in field_definitions:
                        field_definitions[field_name] = {}

                    if val in ('""', "''"):
                        # Looks like "key=''" | 'key=""'
                        val = ''
                    elif not val:  # Looks like 'key='
                        val = None
                    else:
                        for quote in '\'\"':
                            if val.startswith(quote) and val.endswith(quote):
                                val = val.strip(quote)

                    maybe_kwargs = field_opt.split('.', 1)
                    match maybe_kwargs:
                        case ['kwargs']:
                            # Looks like 'field.kwargs={"k": "v", ...}'
                            val = {} if val is None else eval(val)
                            # Yes, I know, it's eval.
                        case ['kwargs', nested]:
                            if nested.isidentifier():
                                # Looks like 'field.kwargs.k=v'
                                kwarg = {nested: val}
                                val = kwarg
                                field_opt = 'kwargs'
                            else:
                                pass
                        case [not_kwargs]:
                            # `field_opt` is valid.
                            pass
                        case _:
                            # Nonsense
                            msg = f"{opt !r} is invalid option syntax."
                            raise CLIUsageError(msg)

                    field_definitions[field_name].update({field_opt: val})

        # Convert values from strings:
        conversion_map = {}
        for param, typ in FieldConfigSpec.__annotations__.items():
            if typ is bool:
                conversion = str_to_bool
            elif type(typ) is TypeAliasType:
                conversion = typ.__value__
            else:
                conversion = typ
            conversion_map[param] = conversion

        for field, params in field_definitions.items():
            for k, v in params.items():
                if k in conversion_map:
                    field_definitions[field][k] = conversion_map[k](v)

        return field_definitions

    def add_arguments(self) -> None:
        '''Equip the parser with all its arguments.'''
        fields_or_tmpl = self.add_mutually_exclusive_group()
        fields_or_tmpl.add_argument(
            '--template', '-t',
            dest='template',
            help=(
                "A curly-brace-delimited format string."
                " Not valid with --fields/-f options."
            ),
            metavar="'TEMPLATE'",
        )

        fields_or_tmpl.add_argument(
            '--fields', '-f',
            action='extend',
            dest='field_order',
            help=(
                "A list of fields to be displayed."
                " Not valid with --template/-t options."
            ),
            metavar=('FIELDNAME1', 'FIELDNAME2'),
            nargs='+',
            # type=ArgFormatter.SplitFirst(','),
        )

        fields_group = self.add_argument_group(
            title="Options for fields",
            description="These options are not valid when using --template/-t."
        )

        fields_group.add_argument(
            '--separator', '-s',
            dest='separator',
            help=(
                "The character used for joining fields."
                " Only valid with --field/-f options."
            ),
            metavar="'FIELD_SEPARATOR'",
            # type=ArgFormatter.ToTuple(length=2),
        )

        fields_group.add_argument(
            '--join-empty', '-j',
            action='store_true',
            dest='join_empty_fields',
            help=(
                "Include empty field contents instead of hiding them."
                " Only valid with --field/-f options."
            ),
        )

        self.add_argument(
            '--options', '-o',
            action='extend',
            dest='field_options',
            help=(
                "Set arbitrary options for discrete Fields using"
                " dot-attribute syntax."
            ),
            metavar=("FIELD1.OPTION='VAL'", "FIELD2.OPTION='VAL'"),
            nargs='+',
        )

        self.add_argument(
            '--refresh', '-r',
            dest='refresh',
            help=(
                "How often in seconds the bar automatically prints itself."
            ),
            type=float,
        )

        self.add_argument(
            '--count', '-n',
            dest='count',
            help=(
                "Print the bar this many times, then exit."
            ),
            metavar='TIMES',
            type=int,
        )

        self.add_argument(
            '--break-lines', '-b',
            action='store_true',
            dest='break_lines',
            help="Use a newline character at the end of every bar line.",
        )

        self.add_argument(
            '--icons', '-i',
            action='extend',
            dest='icon_pairs',
            help="A mapping of Field names to icons.",
            metavar=("FIELDNAME1='ICON1'", "FIELDNAME2='ICON2'"),
            nargs='+',
        )

        self.add_argument(
            '--from-icons',
            action='extend',
            dest='from_icons',
            help=(
                "A mapping of Field names to icons."
                " Use for the Field order as well."
            ),
            metavar=("FIELDNAME1='ICON1'", "FIELDNAME2='ICON2'"),
            nargs='+',
        )

        self.add_argument(
            '--unicode', '-u',
            action='store_true',
            dest='unicode',
            help="Prefer Unicode versions of Field icons, if provided.",
        )

        self.add_argument(
            '--no-unicode', '-U',
            action='store_false',
            dest='unicode',
            help="Prefer ASCII versions of Field icons, if provided.",
        )

        self.add_argument(
            '--config', '-c',
            dest='config_file',
            help=(
                "The config file to use for default settings."
            ),
            metavar='FILE',
        )

        self.add_argument(
            '--dump', '-d',
            action='store_const',
            const=4,
            dest='dump_config',
            help=(
                "Instead of running mybar, print a config file using"
                " options specified in the command."
                # " Optionally pass a number of spaces by which to indent."
            ),
        )

##        self.add_argument(
##            '--dump-raw', '-D',
##            action='store_const',
##            const=None,
##            dest='dump_config',
##            help=(
##                "Instead of running the Bar, print a config file using"
##                " options specified in the command."
##            ),
##        )

        self.add_argument(
            '--debug',
            action='store_true',
            help="Use debug mode. (Not implemented)",
        )

        self.add_argument(
            '--version', '-v',
            action='version',
            version=f"{__package__} {__version__}",
        )

    def quit(self, message: str = "Exiting...") -> Never:
        '''Print a message and exit the program.'''
        self.exit(1, message + '\n')


class HighlightMethod(Enum):
    '''Ways to present default option names.'''
    CAPITALIZE = 'CAPITALIZE'
    # UNDERLINE = 'UNDERLINE'
    # BOLD = 'BOLD'


class OptionsAsker:
    '''
    A tool for presenting options and gathering user input.

    :param opts: A mapping of options
    :type opts: :class:`OptSpec`

    :param default: The default option name
    :type default: :class:`str`

    :param question: Give the user context, defaults to ``""``
    :type question: :class:`str`

    :param case_sensitive: Options are case-sensitive, defaults to ``False``
    :type case_sensitive: :class:`bool`

    :param highlight_method: How to differentiate the default option
    :type highlight_method: :class:`HighlightMethod`

    :raises: :exc:`ValueError` if `default` is not a key in `opts`
    '''
    MISSING = None

    def __init__(
        self,
        opts: OptSpec,
        default: str,
        question: str = "",
        case_sensitive: bool = False,
        highlight_method: HighlightMethod | None = HighlightMethod.CAPITALIZE,
    ) -> None:
        if default not in opts:
            raise ValueError(
                f"The option dict {opts!r} "
                f"must contain the provided default option, {default!r}"
            )
        self.default = default
        self.default_val = opts[default]
        self.question = question
        self.case_sensitive = case_sensitive
        self.choices = opts.copy()

        self.optstrings = self.gen_optstrings(highlight_method)

        self.choices[''] = self.default_val

    def gen_optstrings(
        self,
        highlight_method: HighlightMethod | None = HighlightMethod.CAPITALIZE,
    ) -> tuple[OptName]:
        '''
        A tuple of option names with the default highlighted.

        :param highlight_method: How the default option should be
            differentiated, defaults to :obj:`HighlightMethod.CAPITALIZE`
        :type highlight_method: :class:`HighlightMethod`

        :returns: A tuple of option names
        :rtype: tuple[:class:`OptName`]
        '''
        match highlight_method:
            case HighlightMethod.CAPITALIZE:
                optstrings = (
                    choice.upper() if val == self.default_val # in default
                    else choice
                    for choice, val in self.choices.items()
                )
            case _:
                optstrings = self.choices

        return tuple(optstrings)

    def ask(
        self,
        prompt: str = None,
        repeat_prompt: bool = True,
    ) -> Any:
        '''
        Show a prompt and get user input for an option.
        Block until a valid option is given.

        :param prompt: Shows possible option names for the user to input
        :type prompt: :class:`string`, optional

        :param repeat_prompt: Repeat the whole prompt if the user enters
            an invalid option, defaults to ``True``
            Setting this to ``False`` is useful if
            :attr:`OptionsAsker.question` is very long.

        :type repeat_prompt: :class:`bool`

        :returns: The option value chosen by the user
        :rtype: :class:`Any`
        '''
        answer = self.MISSING
        options = f"[{'/'.join(self.optstrings)}]"
        if prompt is None:
            q = self.question
            prompt = f"{q} {options} " if q else f"{options} "

        prompted = False
        while answer not in self.choices:
            answer = input(prompt)
            if not self.case_sensitive:
                answer = answer.casefold()
            if repeat_prompt or prompted:
                continue
            prompt = options + " "
        return self.choices.get(answer)


def welcome_new_users() -> None:
    msg = (
        f"-- {PROG} --"
    )
    print(msg)


##class ConfigWizard:
class FileManager:

    @classmethod
    def _get_write_new_approval(
        cls,
        file: PathLike,
        dft_choice: bool
    ) -> bool:
        '''
        Get approval to write a new config file using
            :class:`cli.OptionsAsker`.

        :param file: The path to the config file
        :type file: :class:`PathLike`

        :param dft_choice: The default option to present to the user
            (``False`` means do not write)
        :type dft_choice: :class:`bool`
        '''
        question = f"Would you like to write a new config file at {file!r}?"

        absolute = os.path.abspath(os.path.expanduser(CONFIG_FILE))
        if not os.path.exists(absolute):
            maybe_default = ' '
            if file == CONFIG_FILE:
                welcome_new_users()
                cls._maybe_make_config_dir()
                maybe_default = ' default '
            msg = (
                f"The{maybe_default}config file at {file!r} does not exist."
            )
            print(msg)
            question = "Would you like to make it now?"

        write_options = {'y': True, 'n': False}
        default = 'ny'[dft_choice]
        handler = OptionsAsker(write_options, default, question)

        write_new_file_ok = handler.ask()
        return write_new_file_ok

    @staticmethod
    def _maybe_make_config_dir() -> None:
        '''
        Make a '.config' directory if nonexistent.
        '''
        absolute = os.path.abspath(os.path.expanduser(CONFIG_FILE))
        directory = os.path.dirname(absolute)
        if not os.path.exists(directory):
            os.mkdir(directory)
            print(f"Made new directory {directory!r}")

